import os
import asyncio
import aiofiles
from contextlib import contextmanager
from .exceptions import NotEnoughResources, ProcessInitializationError, WorkerNotAvailable
from .environment_resolver import ResolutionImpossibleError
from . import logging
from . import invocationjob
from .exceptions import AlreadyRunning
from .taskspawn import TaskSpawn
from .enums import WorkerPingReply, TaskScheduleStatus, WorkerState, WorkerType, SpawnStatus, InvocationState, InvocationMessageResult
from .worker_messsage_processor import WorkerControlClient
from .net_classes import WorkerResources
from .net_messages.impl.tcp_simple_command_message_processor import TcpCommandMessageProcessor
from .net_messages.impl.clients import CommandJsonMessageClient
from .net_messages.address import AddressChain
from .net_messages.messages import Message
from .net_messages.exceptions import MessageTransferTimeoutError, MessageReceiveTimeoutError, MessageTransferError
from .net_messages.impl.message_haldlers import CommandMessageHandlerBase


from typing import Awaitable, Callable, Dict, List, Optional, Set, Tuple, TYPE_CHECKING
if TYPE_CHECKING:
    from .scheduler import Scheduler


class SchedulerCommandHandler(CommandMessageHandlerBase):
    def __init__(self, scheduler: "Scheduler"):
        super().__init__()
        self.__scheduler = scheduler

    def command_mapping(self) -> Dict[str, Callable[[dict, CommandJsonMessageClient, Message], Awaitable[None]]]:
        return {
            'pulse': self._command_pulse,
            '_pulse3way_': self._command_pulse3way,  # TODO: remove this when handlers are implemented
            # worker-specific
            'worker.ping': self._command_ping,
            'worker.done': self._command_done,
            'worker.dropped': self._command_dropped,
            'worker.hello': self._command_hello,
            'worker.bye': self._command_bye,
            'forward_invocation_message': self._command_forward_invocation_message,
        }

    #
    # commands
    #

    async def _command_ping(self, args: dict, client: CommandJsonMessageClient, original_message: Message):  # 'ping'
        """
        expects keys:
            worker_addr: address of the worker to query
        returns keys:
            state: worker status as seen by scheduler
        """
        # when worker pings scheduler - scheduler returns the state it thinks the worker is in
        addr = args['worker_addr']
        wid = await self.__scheduler.worker_id_from_address(addr)
        if wid is None:
            state = WorkerState.UNKNOWN
        else:
            state = await self.__scheduler.get_worker_state(wid)
        await client.send_message_as_json({'state': state.value})

    async def _command_pulse(self, args: dict, client: CommandJsonMessageClient, original_message: Message):  # 'pulse'
        """
        expects keys:
        returns keys:
            ok: ok is ok
        """
        await client.send_message_as_json({'ok': True})

    async def _command_done(self, args: dict, client: CommandJsonMessageClient, original_message: Message):  # 'done'
        """
        expects keys:
            task: serialized task
            stdout: task's stdout log (str)
            stderr: task's stderr log (str)
        returns keys:
            ok: ok is ok
        """
        task_data = args['task'].encode('latin1')
        task = await invocationjob.InvocationJob.deserialize_async(task_data)

        stdout = args['stdout']
        stderr = args['stderr']
        await self.__scheduler.task_done_reported(task, stdout, stderr)
        await client.send_message_as_json({'ok': True})

    async def _command_dropped(self, args: dict, client: CommandJsonMessageClient, original_message: Message):  # 'dropped'
        """
        expects keys:
        returns keys:
            ok: ok is ok
        """
        task_data = args['task'].encode('latin1')
        task = await invocationjob.InvocationJob.deserialize_async(task_data)

        stdout = args['stdout']
        stderr = args['stderr']
        await self.__scheduler.task_cancel_reported(task, stdout, stderr)
        await client.send_message_as_json({'ok': True})

    async def _command_hello(self, args: dict, client: CommandJsonMessageClient, original_message: Message):  # 'hello'
        """
        worker reports for duty

        expects keys:
            worker_addr: worker address to talk to
            worker_type: worker type enum value
            worker_res: serialized worker resource capabilities
        returns keys:
            db_uid: scheduler's database uid
        """
        addr = args['worker_addr']
        workertype: WorkerType = WorkerType(args['worker_type'])
        res_data = args['worker_res'].encode('latin1')

        worker_hardware: WorkerResources = WorkerResources.deserialize(res_data)
        await self.__scheduler.add_worker(addr, workertype, worker_hardware, assume_active=True)
        await client.send_message_as_json({'db_uid': self.__scheduler.db_uid()})

    async def _command_bye(self, args: dict, client: CommandJsonMessageClient, original_message: Message):  # 'bye'
        """
        worker reports he's quitting

        expects keys:
            worker_addr: worker address to talk to
        returns keys:
            ok: ok is ok
        """
        addr = args['worker_addr']
        await self.__scheduler.worker_stopped(addr)
        await client.send_message_as_json({'ok': True})

    async def _command_pulse3way(self, args: dict, client: CommandJsonMessageClient, original_message: Message):
        """
        TODO: remove this when handlers are implemented
        This command exists for test purposes only
        """
        await client.send_message_as_json({'phase': 1})
        msg2 = await client.receive_message()
        await client.send_message_as_json({'phase': 2})

    # worker task message forwarding
    async def _command_forward_invocation_message(self, args: dict, client: CommandJsonMessageClient, original_message: Message):
        """
        expects keys:
            dst_invoc_id: receiver's invocation id
            src_invoc_id: sender's invocation id
            addressee: address id, where to address message within worker
            message_data_raw: message_data_raw
            addressee_timeout: timeout in seconds of how long to wait for addressee to start receiving
            overall_timeout: overall delivery waiting timeout. since this is just delivery, the only real waiting
                             may actually happen when waiting for addressee to start receiving, so this timeout
                             is very unlikely to ever hit, unless overall comm is disrupted
        returns keys:
            result: str, operation result
        """
        invoc_id = args['dst_invoc_id']
        invoc_state = await self.__scheduler.get_invocation_state(invoc_id)
        if invoc_state != InvocationState.IN_PROGRESS:
            await client.send_message_as_json({
                'result': InvocationMessageResult.ERROR_IID_NOT_RUNNING.value
            })
            return

        address = await self.__scheduler.get_invocation_worker(invoc_id)
        if address is None:
            await client.send_message_as_json({
                'result': InvocationMessageResult.ERROR_BAD_IID.value
            })
            return

        try:
            with WorkerControlClient.get_worker_control_client(address, self.__scheduler.message_processor()) as worker_client:  # type: WorkerControlClient
                result = await worker_client.send_invocation_message(invoc_id,
                                                                     args['addressee'],
                                                                     args['src_invoc_id'],
                                                                     args['message_data_raw'].encode('latin1'),
                                                                     args['addressee_timeout'],
                                                                     args['overall_timeout'])
        except MessageTransferTimeoutError:
            self._logger.error('could not deliver invocation message, timeout')
            result = InvocationMessageResult.ERROR_DELIVERY_TIMEOUT
        except MessageTransferError:
            self._logger.exception('message transfer failed')
            result = InvocationMessageResult.ERROR_TRANSFER_ERROR
        except Exception:
            self._logger.exception('something wend wrong')
            result = InvocationMessageResult.ERROR_UNEXPECTED

        await client.send_message_as_json({
            'result': result.value
        })


class SchedulerExtraCommandHandler(CommandMessageHandlerBase):
    def __init__(self, scheduler: "Scheduler"):
        super().__init__()
        self.__scheduler = scheduler

    def command_mapping(self) -> Dict[str, Callable[[dict, CommandJsonMessageClient, Message], Awaitable[None]]]:
        return {
            'spawn': self.comm_spawn,
            'nodenametoid': self.comm_node_name_to_id,
            'tupdateattribs': self.comm_update_task_attributes,
        }

    async def comm_spawn(self, args: dict, client: CommandJsonMessageClient, original_message: Message):
        """
        spawn a new task

        expects keys:
            task: serialized TaskSpawn
        returns keys:
            status: SpawnStatus value
            task_id: spawned task id or None if no tasks were spawned
        """
        task_data = args['task'].encode('latin1')
        taskspawn: TaskSpawn = TaskSpawn.deserialize(task_data)

        ret: Tuple[SpawnStatus, Optional[int]] = await self.__scheduler.spawn_tasks(taskspawn)
        await client.send_message_as_json({
            'status': ret[0].value,
            'task_id': ret[1]
        })

    async def comm_node_name_to_id(self, args: dict, client: CommandJsonMessageClient, original_message: Message):
        """
        node name to node id if found

        expects keys:
            name: name of the node to find
        returns keys:
            node_ids: list of int, ids of the nodes with given name
        """
        ids = await self.__scheduler.node_name_to_id(args['name'])
        await client.send_message_as_json({
            'node_ids': list(ids)
        })

    async def comm_update_task_attributes(self, args: dict, client: CommandJsonMessageClient, original_message: Message):
        """
        update task attributes

        expects keys:
            task_id: id of the task to update attributes of
            attribs_to_update: dict of attribute names to attribute values
            attribs_to_delete: list of attribute names to delete
        returns keys:
            ok: ok is ok
        """
        task_id = args['task_id']
        attribs_to_update = args['attribs_to_update']
        attribs_to_delete = set(args['attribs_to_delete'])
        await self.__scheduler.update_task_attributes(task_id, attribs_to_update, attribs_to_delete)
        await client.send_message_as_json({
            'ok': True
        })


class SchedulerMessageProcessor(TcpCommandMessageProcessor):
    def __init__(self, scheduler: "Scheduler", listening_address: Tuple[str, int], *, backlog=4096, connection_pool_cache_time=300):
        super().__init__(listening_address,
                         backlog=backlog,
                         connection_pool_cache_time=connection_pool_cache_time,
                         message_handlers=(SchedulerCommandHandler(scheduler),
                                           SchedulerExtraCommandHandler(scheduler)))
        self.__logger = logging.get_logger('scheduler.message_processor')


#
# Client
#


class SchedulerBaseClient:
    def __init__(self, client: CommandJsonMessageClient):
        self.__client = client

    @classmethod
    @contextmanager
    def get_scheduler_control_client(cls, scheduler_address: AddressChain, processor: TcpCommandMessageProcessor) -> "SchedulerBaseClient":
        with processor.message_client(scheduler_address) as message_client:
            yield SchedulerBaseClient(message_client)

    async def pulse(self):
        await self.__client.send_command('pulse', {})
        reply = await self.__client.receive_message()
        assert (await reply.message_body_as_json()).get('ok', False), 'something is not ok'


class SchedulerWorkerControlClient(SchedulerBaseClient):
    def __init__(self, client: CommandJsonMessageClient):
        super().__init__(client)
        self.__client = client

    @classmethod
    @contextmanager
    def get_scheduler_control_client(cls, scheduler_address: AddressChain, processor: TcpCommandMessageProcessor) -> "SchedulerWorkerControlClient":
        with processor.message_client(scheduler_address) as message_client:
            yield SchedulerWorkerControlClient(message_client)

    async def ping(self, addr: AddressChain) -> WorkerState:
        await self.__client.send_command('worker.ping', {
            'worker_addr': str(addr)
        })
        reply = await self.__client.receive_message()
        return WorkerState((await reply.message_body_as_json())['state'])

    async def report_task_done(self, task: invocationjob.InvocationJob, stdout_file: str, stderr_file: str):
        async with aiofiles.open(stdout_file, 'r') as f:
            stdout = await f.read()
        async with aiofiles.open(stderr_file, 'r') as f:
            stderr = await f.read()
        await self.__client.send_command('worker.done', {
            'task': (await task.serialize_async()).decode('latin1'),
            'stdout': stdout,
            'stderr': stderr
        })
        reply = await self.__client.receive_message()
        assert (await reply.message_body_as_json()).get('ok', False), 'something is not ok'

    async def report_task_canceled(self, task: invocationjob.InvocationJob, stdout_file: str, stderr_file: str):
        async with aiofiles.open(stdout_file, 'r') as f:
            stdout = await f.read()
        async with aiofiles.open(stderr_file, 'r') as f:
            stderr = await f.read()
        await self.__client.send_command('worker.dropped', {
            'task': (await task.serialize_async()).decode('latin1'),
            'stdout': stdout,
            'stderr': stderr
        })
        reply = await self.__client.receive_message()
        assert (await reply.message_body_as_json()).get('ok', False), 'something is not ok'

    async def say_hello(self, address_to_advertise: AddressChain, worker_type: WorkerType, worker_resources: WorkerResources) -> int:
        await self.__client.send_command('worker.hello', {
            'worker_addr': str(address_to_advertise),
            'worker_type': worker_type.value,
            'worker_res': worker_resources.serialize().decode('latin1')
        })
        reply = await self.__client.receive_message()
        return (await reply.message_body_as_json())['db_uid']

    async def say_bye(self, address_of_worker: str):
        await self.__client.send_command('worker.bye', {
            'worker_addr': str(address_of_worker)
        })
        reply = await self.__client.receive_message()
        assert (await reply.message_body_as_json()).get('ok', False), 'something is not ok'


class SchedulerExtraControlClient(SchedulerBaseClient):
    def __init__(self, client: CommandJsonMessageClient):
        super().__init__(client)
        self.__client = client

    @classmethod
    @contextmanager
    def get_scheduler_control_client(cls, scheduler_address: AddressChain, processor: TcpCommandMessageProcessor) -> "SchedulerExtraControlClient":
        with processor.message_client(scheduler_address) as message_client:
            yield SchedulerExtraControlClient(message_client)

    async def spawn(self, task_spawn: TaskSpawn) -> Tuple[SpawnStatus, Optional[int]]:
        await self.__client.send_command('spawn', {
            'task': task_spawn.serialize().decode('latin1')
        })
        reply = await self.__client.receive_message()
        ret_data = await reply.message_body_as_json()
        return SpawnStatus(ret_data['status']), ret_data['task_id']

    async def node_name_to_id(self, name: str) -> List[int]:
        await self.__client.send_command('nodenametoid', {
            'name': name
        })
        reply = await self.__client.receive_message()
        ret_data = await reply.message_body_as_json()
        return list(ret_data['node_ids'])

    async def update_task_attributes(self, task_id: int, attribs_to_update: dict, attribs_to_delete: Set[str]):
        await self.__client.send_command('tupdateattribs', {
            'task_id': task_id,
            'attribs_to_update': attribs_to_update,
            'attribs_to_delete': list(attribs_to_delete),
        })
        reply = await self.__client.receive_message()
        assert (await reply.message_body_as_json()).get('ok')


class SchedulerInvocationMessageClient:
    def __init__(self, client: CommandJsonMessageClient):
        self.__client = client

    @classmethod
    @contextmanager
    def get_scheduler_control_client(cls, scheduler_address: AddressChain, processor: TcpCommandMessageProcessor) -> "SchedulerInvocationMessageClient":
        with processor.message_client(scheduler_address) as message_client:
            yield SchedulerInvocationMessageClient(message_client)

    async def send_invocation_message(self,
                                      destination_invocation_id: int,
                                      destination_addressee: str,
                                      source_invocation_id: Optional[int],
                                      message_body: bytes,
                                      *,
                                      addressee_timeout: float = 90,
                                      overall_timeout: float = 300) -> InvocationMessageResult:
        if overall_timeout < addressee_timeout:
            overall_timeout = addressee_timeout

        await self.__client.send_command('forward_invocation_message', {
            'dst_invoc_id': destination_invocation_id,
            'src_invoc_id': source_invocation_id,
            'addressee': destination_addressee,
            'addressee_timeout': addressee_timeout,
            'overall_timeout': overall_timeout,
            'message_data_raw': message_body.decode('latin1'),
        })
        try:
            return InvocationMessageResult((await (await self.__client.receive_message(timeout=overall_timeout)).message_body_as_json())['result'])
        except MessageReceiveTimeoutError:
            return InvocationMessageResult.ERROR_DELIVERY_TIMEOUT



import asyncio
from dataclasses import dataclass
from uuid import UUID
from .messages import Message

from typing import Dict


@dataclass
class RefCountQueue:
    queue: asyncio.Queue
    ref_count: int = 0

    def __enter__(self):
        self.ref_count += 1
        return self.queue

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.ref_count -= 1


class MessageQueue:
    def __init__(self):
        self.__session_map: Dict[UUID, RefCountQueue] = {}

    async def put_message(self, message: Message):
        session = message.message_session()
        if session not in self.__session_map:
            self.__session_map[session] = RefCountQueue(asyncio.Queue())
        with self.__session_map[session] as queue:
            await queue.put(message)

    async def get_message(self, session):
        if session not in self.__session_map:
            self.__session_map[session] = RefCountQueue(asyncio.Queue())

        with self.__session_map[session] as queue:
            message = await queue.get()
            if self.__session_map[session].ref_count == 1:  # 1 means it's just us
                self.__session_map.pop(session)
            return message



import os
import fnmatch
import inspect
from lifeblood.taskspawn import TaskSpawn
from lifeblood.basenode import BaseNodeWithTaskRequirements
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.invocationjob import InvocationJob
from lifeblood.processingcontext import ProcessingContext
from lifeblood.uidata import NodeParameterType
from lifeblood.enums import WorkerType

from typing import Iterable


def node_class():
    return FilePattern


def _scan_one_level(basepath, levels):
    files = []
    if len(levels) == 0:
        return [basepath]
    level = levels[0]
    if level == '':
        return _scan_one_level(basepath, levels[1:])
    if '*' not in level:  # constant value, not a pattern
        path = os.path.join(basepath, level)
        if not os.path.exists(path):
            return []
        return _scan_one_level(path, levels[1:])
    else:
        if not os.path.isdir(basepath):
            return []
        for file_name in os.listdir(basepath):
            if not fnmatch.fnmatch(file_name, level):
                continue
            path = os.path.join(basepath, file_name)
            if not os.path.exists(path):
                continue
            files.extend(_scan_one_level(path, levels[1:]))
    return files


class FilePattern(BaseNodeWithTaskRequirements):
    def __init__(self, name):
        super(FilePattern, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.24, 0.25, 0.48)
            ui.add_parameter('on workers', 'submit to workers', NodeParameterType.BOOL, False)
            ui.add_parameter('pattern', 'file pattern', NodeParameterType.STRING, '')
            ui.parameter('worker type').set_value(WorkerType.SCHEDULER_HELPER.value)

    @classmethod
    def label(cls) -> str:
        return 'file pattern'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'file', 'pattern', 'list', 'ls'

    @classmethod
    def type_name(cls) -> str:
        return 'filepattern'

    @classmethod
    def description(cls) -> str:
        return 'scans file system according to given pattern, like /some/path/dir_*_smth/imagefile.*.exr\n' \
               'if "on worker" is checked - worker will perform the scan,  \n' \
               'otherwise scan will be done by scheduler.\n' \
               '\n' \
               'if you scan thousands of files it\'s more optimal to do work "on worker"  \n' \
               'as it will produce new items asynchronously,  \n' \
               'while when not "on worker" - scan happens synchronously\n'

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        pattern = context.param_value('pattern')
        pattern = pattern.replace('\\', '/')  # simpler to just work with /  no point working with OS-dependant ones
        if context.param_value('on workers'):
            script = inspect.getsource(_scan_one_level)
            script += '\n\n' \
                      "parts = pattern.split('/')\n" \
                      "if len(parts):\n" \
                      "    return ProcessingResult()\n" \
                      "files = _scan_one_level(parts[0] or '/', parts[1:])\n"

            job = InvocationJob(['python', ':/script.py'])
            job.set_extra_file('script.py', script)
            return ProcessingResult(job)
        else:
            parts = pattern.split('/')
            if len(parts):
                return ProcessingResult()
            files = _scan_one_level(parts[0] or '/', parts[1:])
            res = ProcessingResult()
            for i, file in enumerate(files):
                attrs = {'file': file}
                res.add_spawned_task(TaskSpawn(f"{context.task_field('name')}_file{i}", task_attributes=attrs))

        return res

    def postprocess_task(self, context: ProcessingContext) -> ProcessingResult:
        return ProcessingResult()



from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.processingcontext import ProcessingContext
from lifeblood.enums import NodeParameterType
from lifeblood.uidata import NodeUi, Parameter, VerticalParametersLayout, ParameterHierarchyItem, ParametersLayoutBase
from lifeblood.node_visualization_classes import NodeColorScheme

from typing import Iterable


def node_class():
    return SwitchTasks


class SwitchTasks(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'switch tasks'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'switch', 'core'

    @classmethod
    def type_name(cls) -> str:
        return 'swtich'

    def __init__(self, name: str):
        super(SwitchTasks, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_parameter_to_control_output_count('outputs count', 'number of outputs')
            ui.add_parameter('output', 'select output', NodeParameterType.INT, 0, can_have_expressions=True)

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        switch = context.param_value('output')
        outputs = self.get_ui().outputs_names()

        res = ProcessingResult()
        res.set_node_output_name(outputs[min(len(outputs) - 1, max(0, switch))])
        return res



import sys
import os
from pathlib import Path
import time
from datetime import datetime
import json
import itertools
import asyncio
import aiosqlite
import aiofiles
from aiorwlock import RWLock
from contextlib import asynccontextmanager

from .. import logging
from .. import paths
#from ..worker_task_protocol import WorkerTaskClient
from ..worker_messsage_processor import WorkerControlClient
from ..scheduler_task_protocol import SchedulerTaskProtocol, SpawnStatus
from ..scheduler_ui_protocol import SchedulerUiProtocol
from ..invocationjob import InvocationJob
from ..environment_resolver import EnvironmentResolverArguments
from ..broadcasting import create_broadcaster
from ..simple_worker_pool import WorkerPool
from ..nethelpers import address_to_ip_port, get_default_addr, get_default_broadcast_addr
from ..net_classes import WorkerResources
from ..taskspawn import TaskSpawn
from ..basenode import BaseNode
from ..exceptions import *
from .. import pluginloader
from ..enums import WorkerState, WorkerPingState, TaskState, InvocationState, WorkerType, \
    SchedulerMode, TaskGroupArchivedState
from ..config import get_config
from ..misc import atimeit, alocking
from ..defaults import scheduler_port as default_scheduler_port, ui_port as default_ui_port, scheduler_message_port as default_scheduler_message_port
from .. import aiosqlite_overlay
from ..ui_protocol_data import TaskData, TaskDelta, IncompleteInvocationLogData, InvocationLogData

from ..net_messages.address import DirectAddress, AddressChain
from ..scheduler_message_processor import SchedulerMessageProcessor

from .data_access import DataAccess
from .scheduler_component_base import SchedulerComponentBase
from .pinger import Pinger
from .task_processor import TaskProcessor
from .ui_state_accessor import UIStateAccessor

from typing import Optional, Any, Tuple, List, Iterable, Union, Dict


class Scheduler:
    def __init__(self, db_file_path, *, do_broadcasting: Optional[bool] = None, broadcast_interval: Optional[int] = None,
                 helpers_minimal_idle_to_ensure=1,
                 server_addr: Optional[Tuple[str, int, int]] = None,
                 server_ui_addr: Optional[Tuple[str, int]] = None):
        """
        TODO: add a docstring

        :param db_file_path:
        :param do_broadcasting:
        :param helpers_minimal_idle_to_ensure:
        :param server_addr:
        :param server_ui_addr:
        """
        self.__logger = logging.get_logger('scheduler')
        self.__logger.info('loading core plugins')
        pluginloader.init()  # TODO: move it outside of constructor
        self.__node_objects: Dict[int, BaseNode] = {}
        self.__node_objects_locks: Dict[int, RWLock] = {}
        config = get_config('scheduler')

        # this lock will prevent tasks from being reported cancelled and done at the same exact time should that ever happen
        # this lock is overkill already, but we can make it even more overkill by using set of locks for each invoc id
        # which would be completely useless now cuz sqlite locks DB as a whole, not even a single table, especially not just parts of table
        self.__invocation_reporting_lock = asyncio.Lock()

        self.__all_components = None
        self.__started_event = asyncio.Event()
        
        loop = asyncio.get_event_loop()

        if db_file_path is None:
            config = get_config('scheduler')
            db_file_path = config.get_option_noasync('core.database.path', str(paths.default_main_database_location()))
        if not db_file_path.startswith('file:'):  # if schema is used - we do not modify the db uri in any way
            db_file_path = os.path.realpath(os.path.expanduser(db_file_path))

        self.__logger.debug(f'starting scheduler with database: {db_file_path}')

        self.db_path = db_file_path
        self.data_access = DataAccess(db_file_path, 30)
        ##

        self.__use_external_log = config.get_option_noasync('core.database.store_logs_externally', False)
        self.__external_log_location: Optional[Path] = config.get_option_noasync('core.database.store_logs_externally_location', None)
        if self.__use_external_log and not self.__external_log_location:
            raise SchedulerConfigurationError('if store_logs_externally is set - store_logs_externally_location must be set too')
        if self.__use_external_log:
            external_log_path = Path(self.__use_external_log)
            if external_log_path.exists() and external_log_path.is_file():
                external_log_path.unlink()
            if not external_log_path.exists():
                external_log_path.mkdir(parents=True)
            if not os.access(self.__external_log_location, os.X_OK | os.W_OK):
                raise RuntimeError('cannot write to external log location provided')

        self.__pinger: Pinger = Pinger(self)
        self.__task_processor: TaskProcessor = TaskProcessor(self)
        self.ui_state_access: UIStateAccessor = UIStateAccessor(self)

        server_ip = None
        if server_addr is None:
            server_ip = config.get_option_noasync('core.server_ip', get_default_addr())
            server_port = config.get_option_noasync('core.server_port', default_scheduler_port())
            message_server_port = config.get_option_noasync('core.server_message_port', default_scheduler_message_port())
        else:
            server_ip, server_port, message_server_port = server_addr
        if server_ui_addr is None:
            ui_ip = config.get_option_noasync('core.ui_ip', server_ip or get_default_addr())
            ui_port = config.get_option_noasync('core.ui_port', default_ui_port())
        else:
            ui_ip, ui_port = server_ui_addr
        self.__stop_event = asyncio.Event()
        self.__server_closing_task = None
        self.__cleanup_tasks = None

        self.__server = None
        self.__server_coro = loop.create_server(self._scheduler_protocol_factory, server_ip, server_port, backlog=16)
        self.__server_address = ':'.join((server_ip, str(server_port)))
        self.__message_processor: Optional[SchedulerMessageProcessor] = None
        self.__message_address: Tuple[str, int] = (server_ip, message_server_port)
        self.__ui_server = None
        self.__ui_server_coro = loop.create_server(self._ui_protocol_factory, ui_ip, ui_port, backlog=16)
        self.__ui_address = ':'.join((ui_ip, str(ui_port)))
        if do_broadcasting is None:
            do_broadcasting = config.get_option_noasync('core.broadcast', True)
        if do_broadcasting:
            if broadcast_interval is None or broadcast_interval <= 0:
                broadcast_interval = config.get_option_noasync('core.broadcast_interval', 10)
            broadcast_info = json.dumps({
                'message_address': str(DirectAddress.from_host_port(*self.__message_address)),
                'worker': self.__server_address,
                'ui': self.__ui_address
            })
            self.__broadcasting_server = None
            self.__broadcasting_server_coro = create_broadcaster('lifeblood_scheduler', broadcast_info, ip=get_default_broadcast_addr(), broadcast_interval=broadcast_interval)
        else:
            self.__broadcasting_server = None
            self.__broadcasting_server_coro = None

        self.__worker_pool = None
        self.__worker_pool_helpers_minimal_idle_to_ensure = helpers_minimal_idle_to_ensure

        self.__event_loop = asyncio.get_running_loop()
        assert self.__event_loop is not None, 'Scheduler MUST be created within working event loop, in the main thread'

    def get_event_loop(self):
        return self.__event_loop

    def _scheduler_protocol_factory(self):
        return SchedulerTaskProtocol(self)

    def _ui_protocol_factory(self):
        return SchedulerUiProtocol(self)

    def db_uid(self) -> int:
        """
        unique id that was generated on creation for the DB currently in use

        :return: 64 bit unsigned int
        """
        return self.data_access.db_uid

    def wake(self):
        """
        scheduler may go into DORMANT mode when he things there's nothing to do
        in that case wake() call exits DORMANT mode immediately
        if wake is not called on some change- eventually scheduler will check it's shit and will decide to exit DORMANT mode on it's own, it will just waste some time first
        if currently not in DORMANT mode - nothing will happen

        :return:
        """
        self.__task_processor.wake()
        self.__pinger.wake()

    def poke_task_processor(self):
        """
        kick that lazy ass to stop it's waitings and immediately perform another processing iteration
        this is not connected to wake, __sleep and DORMANT mode,
        this is just one-time kick
        good to perform when task was changed somewhere async, outside of task_processor

        :return:
        """
        self.__task_processor.poke()

    def _component_changed_mode(self, component: SchedulerComponentBase, mode: SchedulerMode):
        if component == self.__task_processor and mode == SchedulerMode.DORMANT:
            self.__logger.info('task processor switched to DORMANT mode')
            self.__pinger.sleep()

    def message_processor(self) -> SchedulerMessageProcessor:
        """
        get scheduler's main message processor
        """
        return self.__message_processor

    async def get_node_type_and_name_by_id(self, node_id: int) -> (str, str):
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            async with con.execute('SELECT "type", "name" FROM "nodes" WHERE "id" = ?', (node_id,)) as nodecur:
                node_row = await nodecur.fetchone()
        if node_row is None:
            raise RuntimeError(f'node with given id {node_id} does not exist')
        return node_row['type'], node_row['name']

    @asynccontextmanager
    async def node_object_by_id_for_reading(self, node_id: int):
        async with self.get_node_lock_by_id(node_id).reader_lock:
            yield await self._get_node_object_by_id(node_id)

    @asynccontextmanager
    async def node_object_by_id_for_writing(self, node_id: int):
        async with self.get_node_lock_by_id(node_id).writer_lock:
            yield await self._get_node_object_by_id(node_id)

    async def _get_node_object_by_id(self, node_id: int) -> BaseNode:
        """
        When accessing node this way - be aware that you SHOULD ensure your access happens within a lock
        returned by get_node_lock_by_id.
        If you don't want to deal with that - use scheduler's wrappers to access nodes in a safe way
        (lol, wrappers are not implemented)

        :param node_id:
        :return:
        """
        if node_id in self.__node_objects:
            return self.__node_objects[node_id]
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            async with con.execute('SELECT * FROM "nodes" WHERE "id" = ?', (node_id,)) as nodecur:
                node_row = await nodecur.fetchone()
            if node_row is None:
                raise RuntimeError('node id is invalid')

            node_type = node_row['type']
            if node_type not in pluginloader.plugins:
                raise RuntimeError('node type is unsupported')

            if node_row['node_object'] is not None:
                self.__node_objects[node_id] = await BaseNode.deserialize_async(node_row['node_object'], self, node_id)
                return self.__node_objects[node_id]

            # newnode: BaseNode = pluginloader.plugins[node_type].create_node_object(node_row['name'], self)
            newnode: BaseNode = pluginloader.create_node(node_type, node_row['name'], self, node_id)
            self.__node_objects[node_id] = newnode
            await con.execute('UPDATE "nodes" SET node_object = ? WHERE "id" = ?',
                              (await newnode.serialize_async(), node_id))
            await con.commit()

            return newnode

    def get_node_lock_by_id(self, node_id: int) -> RWLock:
        """
        All read/write operations for a node should be locked within a per node rw lock that scheduler maintains.
        Usually you do NOT have to be concerned with this.
        But in cases you get the node object with functions like get_node_object_by_id.
        it is your responsibility to ensure data is locked when accessed.
        Lock is not part of the node itself.

        :param node_id: node id to get lock to
        :return: rw lock for the node
        """
        if node_id not in self.__node_objects_locks:
            self.__node_objects_locks[node_id] = RWLock(fast=True)  # read about fast on github. the points is if we have awaits inside critical section - it's safe to use fast
        return self.__node_objects_locks[node_id]

    async def get_task_attributes(self, task_id: int) -> Tuple[Dict[str, Any], Optional[EnvironmentResolverArguments]]:
        """
        get tasks, atributes and it's enviroment resolver's attributes

        :param task_id:
        :return:
        """
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            async with con.execute('SELECT attributes, environment_resolver_data FROM tasks WHERE "id" = ?', (task_id,)) as cur:
                res = await cur.fetchone()
            if res is None:
                raise RuntimeError('task with specified id was not found')
            env_res_args = None
            if res['environment_resolver_data'] is not None:
                env_res_args = await EnvironmentResolverArguments.deserialize_async(res['environment_resolver_data'])
            return await asyncio.get_event_loop().run_in_executor(None, json.loads, res['attributes']), env_res_args

    async def get_task_fields(self, task_id: int) -> Dict[str, Any]:
        """
        returns information about the given task, excluding thicc fields like attributes or env resolver
        for those - use get_task_attributes

        :param task_id:
        :return:
        """
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            async with con.execute('SELECT "id", "name", parent_id, children_count, active_children_count, "state", paused, '
                                   '"node_id", split_level, priority, "dead" FROM tasks WHERE "id" == ?', (task_id,)) as cur:
                res = await cur.fetchone()
            if res is None:
                raise RuntimeError('task with specified id was not found')
            return dict(res)

    async def task_name_to_id(self, name: str) -> List[int]:
        """
        get the list of task ids that have specified name

        :param name:
        :return:
        """
        async with self.data_access.data_connection() as con:
            async with con.execute('SELECT "id" FROM "tasks" WHERE "name" = ?', (name,)) as cur:
                return list(x[0] for x in await cur.fetchall())

    async def get_task_invocation_serialized(self, task_id: int) -> Optional[bytes]:
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            async with con.execute('SELECT work_data FROM tasks WHERE "id" = ?', (task_id,)) as cur:
                res = await cur.fetchone()
            if res is None:
                raise RuntimeError('task with specified id was not found')
            return res[0]

    async def worker_id_from_address(self, addr: str) -> Optional[int]:
        async with self.data_access.data_connection() as con:
            async with con.execute('SELECT "id" FROM workers WHERE last_address = ?', (addr,)) as cur:
                ret = await cur.fetchone()
        if ret is None:
            return None
        return ret[0]

    async def get_worker_state(self, wid: int, con: Optional[aiosqlite.Connection] = None) -> WorkerState:
        if con is None:
            async with self.data_access.data_connection() as con:
                async with con.execute('SELECT "state" FROM "workers" WHERE "id" = ?', (wid,)) as cur:
                    res = await cur.fetchone()
        else:
            async with con.execute('SELECT "state" FROM "workers" WHERE "id" = ?', (wid,)) as cur:
                res = await cur.fetchone()
        if res is None:
            raise ValueError(f'worker with given wid={wid} was not found')
        return WorkerState(res[0])

    async def get_task_invocation(self, task_id: int):
        data = await self.get_task_invocation_serialized(task_id)
        if data is None:
            return None
        return await InvocationJob.deserialize_async(data)

    async def get_invocation_worker(self, invocation_id: int) -> Optional[AddressChain]:
        async with self.data_access.data_connection() as con:
            async with con.execute(
                    'SELECT workers.last_address '
                    'FROM invocations LEFT JOIN workers '
                    'ON invocations.worker_id == workers.id '
                    'WHERE invocations.id == ?', (invocation_id,)) as cur:
                res = await cur.fetchone()
        if res is None:
            return None
        return AddressChain(res[0])

    async def get_invocation_state(self, invocation_id: int) -> Optional[InvocationState]:
        async with self.data_access.data_connection() as con:
            async with con.execute(
                    'SELECT state FROM invocations WHERE id == ?', (invocation_id,)) as cur:
                res = await cur.fetchone()
        if res is None:
            return None
        return InvocationState(res[0])

    def stop(self):
        async def _server_closer():
            # ensure all components stop first
            await self.__pinger.wait_till_stops()
            await self.__task_processor.wait_till_stops()
            await self.__worker_pool.wait_till_stops()
            await self.__ui_server.wait_closed()
            if self.__server is not None:
                self.__server.close()
                await self.__server.wait_closed()
            self.__logger.debug('stopping message processor...')
            self.__message_processor.stop()
            await self.__message_processor.wait_till_stops()
            self.__logger.debug('message processor stopped')

        async def _db_cache_writeback():
            await self.__pinger.wait_till_stops()
            await self.__task_processor.wait_till_stops()
            await self.__server_closing_task
            await self._save_all_cached_nodes_to_db()
            await self.data_access.write_back_cache()

        if self.__stop_event.is_set():
            self.__logger.error('cannot double stop!')
            return  # no double stopping
        if not self.__started_event.is_set():
            self.__logger.error('cannot stop what is not started!')
            return
        self.__logger.info('STOPPING SCHEDULER')
        self.__stop_event.set()  # this will stop things including task_processor
        self.__pinger.stop()
        self.__task_processor.stop()
        self.ui_state_access.stop()
        self.__worker_pool.stop()
        self.__server_closing_task = asyncio.create_task(_server_closer())  # we ensure worker pool stops BEFORE server, so workers have chance to report back
        self.__cleanup_tasks = [asyncio.create_task(_db_cache_writeback())]
        if self.__ui_server is not None:
            self.__ui_server.close()

    def _stop_event_wait(self):  # TODO: this is currently being used by ui proto to stop long connections, but not used in task proto, but what if it'll also get long living connections?
        return self.__stop_event.wait()

    async def start(self):
        # prepare
        async with self.data_access.data_connection() as con:
            # we play it the safest for now:
            # all workers set to UNKNOWN state, all active invocations are reset, all tasks in the middle of processing are reset to closest waiting state
            con.row_factory = aiosqlite.Row
            await con.execute('UPDATE "tasks" SET "state" = ? WHERE "state" IN (?, ?)',
                              (TaskState.READY.value, TaskState.IN_PROGRESS.value, TaskState.INVOKING.value))
            await con.execute('UPDATE "tasks" SET "state" = ? WHERE "state" = ?',
                              (TaskState.WAITING.value, TaskState.GENERATING.value))
            await con.execute('UPDATE "tasks" SET "state" = ? WHERE "state" = ?',
                              (TaskState.POST_WAITING.value, TaskState.POST_GENERATING.value))
            await con.execute('UPDATE "invocations" SET "state" = ? WHERE "state" = ?', (InvocationState.FINISHED.value, InvocationState.IN_PROGRESS.value))
            await con.execute('UPDATE workers SET "ping_state" = ?', (WorkerPingState.UNKNOWN.value,))
            await con.execute('UPDATE "workers" SET "state" = ?', (WorkerState.UNKNOWN.value,))
            await con.commit()

            # update volatile mem cache:
            async with con.execute('SELECT "id", last_seen, last_checked, ping_state FROM workers') as worcur:
                async for row in worcur:
                    self.data_access.mem_cache_workers_state[row['id']] = {k: row[k] for k in dict(row)}

        # start
        self.__server = await self.__server_coro
        self.__ui_server = await self.__ui_server_coro
        # start message processor
        self.__message_processor = SchedulerMessageProcessor(self, self.__message_address)
        await self.__message_processor.start()
        self.__worker_pool = WorkerPool(WorkerType.SCHEDULER_HELPER,
                                        minimal_idle_to_ensure=self.__worker_pool_helpers_minimal_idle_to_ensure,
                                        scheduler_address=self.server_message_address())
        await self.__worker_pool.start()
        #
        if self.__broadcasting_server_coro is not None:
            self.__broadcasting_server = await self.__broadcasting_server_coro
        await self.__task_processor.start()
        await self.__pinger.start()
        await self.ui_state_access.start()
        # run
        self.__all_components = \
              asyncio.gather(self.__task_processor.wait_till_stops(),
                             self.__pinger.wait_till_stops(),
                             self.ui_state_access.wait_till_stops(),
                             self.__server.wait_closed(),  # TODO: shit being waited here below is very unnecessary
                             self.__ui_server.wait_closed(),
                             self.__worker_pool.wait_till_stops())

        self.__started_event.set()

    async def wait_till_starts(self):
        return await self.__started_event.wait()

    async def wait_till_stops(self):
        await self.__started_event.wait()
        assert self.__all_components is not None
        await self.__all_components
        await self.__server_closing_task
        for task in self.__cleanup_tasks:
            await task

    async def _save_all_cached_nodes_to_db(self):
        self.__logger.info('saving nodes to db')
        for node_id in self.__node_objects:
            await self.save_node_to_database(node_id)
            self.__logger.debug(f'node {node_id} saved to db')

    def is_started(self):
        return self.__started_event.is_set()

    def is_stopping(self) -> bool:
        """
        True if stopped or in process of stopping
        """
        return self.__stop_event.is_set()

    #
    # helper functions
    #

    async def reset_invocations_for_worker(self, worker_id: int, con: aiosqlite_overlay.ConnectionWithCallbacks, also_update_resources=True) -> bool:
        """

        :param worker_id:
        :param con:
        :param also_update_resources:
        :return: need commit?
        """
        async with con.execute('SELECT * FROM invocations WHERE "worker_id" = ? AND "state" == ?',
                               (worker_id, InvocationState.IN_PROGRESS.value)) as incur:
            all_invoc_rows = await incur.fetchall()  # we don't really want to update db while reading it
        need_commit = False
        for invoc_row in all_invoc_rows:  # mark all (probably single one) invocations
            need_commit = True
            self.__logger.debug("fixing dangling invocation %d" % (invoc_row['id'],))
            await con.execute('UPDATE invocations SET "state" = ? WHERE "id" = ?',
                              (InvocationState.FINISHED.value, invoc_row['id']))
            await con.execute('UPDATE tasks SET "state" = ? WHERE "id" = ?',
                              (TaskState.READY.value, invoc_row['task_id']))
            con.add_after_commit_callback(self.ui_state_access.scheduler_reports_task_updated, invoc_row['task_id'])  # ui event
        if also_update_resources:
            need_commit = need_commit or await self._update_worker_resouce_usage(worker_id, connection=con)
        return need_commit

    #
    # invocation consistency checker
    async def invocation_consistency_checker(self):
        """
        both scheduler and woker might crash at any time. so we need to check that
        worker may crash working on a task (
        :return:
        """
        pass

    #
    # callbacks

    #
    # worker reports done task
    async def task_done_reported(self, task: InvocationJob, stdout: str, stderr: str):
        async with self.__invocation_reporting_lock, \
                   self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            self.__logger.debug('task finished reported %s code %s', repr(task), task.exit_code())
            # sanity check
            async with con.execute('SELECT "state" FROM invocations WHERE "id" = ?', (task.invocation_id(),)) as cur:
                invoc = await cur.fetchone()
                if invoc is None:
                    self.__logger.error('reported task has non existing invocation id %d' % task.invocation_id())
                    return
                if invoc['state'] != InvocationState.IN_PROGRESS.value:
                    self.__logger.warning('reported task for a finished invocation. assuming that worker failed to cancel task previously and ignoring invocation results.')
                    return
            await con.execute('UPDATE invocations SET "state" = ?, "return_code" = ?, "runtime" = ? WHERE "id" = ?',
                              (InvocationState.FINISHED.value, task.exit_code(), task.running_time(), task.invocation_id()))
            async with con.execute('SELECT * FROM invocations WHERE "id" = ?', (task.invocation_id(),)) as incur:
                invocation = await incur.fetchone()
            assert invocation is not None

            await con.execute('UPDATE workers SET "state" = ? WHERE "id" = ?',
                              (WorkerState.IDLE.value, invocation['worker_id']))
            await self._update_worker_resouce_usage(invocation['worker_id'], connection=con)  # remove resource usage info
            tasks_to_wait = []
            if not self.__use_external_log:
                await con.execute('UPDATE invocations SET "stdout" = ?, "stderr" = ? WHERE "id" = ?',
                                  (stdout, stderr, task.invocation_id()))
            else:
                await con.execute('UPDATE invocations SET "log_external" = 1 WHERE "id" = ?',
                                  (task.invocation_id(),))
                tasks_to_wait.append(asyncio.create_task(self._save_external_logs(task.invocation_id(), stdout, stderr)))

            if task.invocation_id() in self.data_access.mem_cache_invocations:
                del self.data_access.mem_cache_invocations[task.invocation_id()]

            ui_task_delta = TaskDelta(invocation['task_id'])  # for ui event
            if task.finished_needs_retry():  # max retry count will be checked by task processor
                await con.execute('UPDATE tasks SET "state" = ? WHERE "id" = ?',
                                  (TaskState.READY.value, invocation['task_id']))
                ui_task_delta.state = TaskState.READY  # for ui event
            elif task.finished_with_error():
                state_details = json.dumps({'message': f'see invocation #{invocation["id"]} log for details',
                                            'happened_at': TaskState.IN_PROGRESS.value,
                                            'type': 'invocation'})
                await con.execute('UPDATE tasks SET "state" = ?, "state_details" = ? WHERE "id" = ?',
                                  (TaskState.ERROR.value,
                                   state_details,
                                   invocation['task_id']))
                ui_task_delta.state = TaskState.ERROR  # for ui event
                ui_task_delta.state_details = state_details  # for ui event
            else:
                await con.execute('UPDATE tasks SET "state" = ? WHERE "id" = ?',
                                  (TaskState.POST_WAITING.value, invocation['task_id']))
                ui_task_delta.state = TaskState.POST_WAITING  # for ui event

            con.add_after_commit_callback(self.ui_state_access.scheduler_reports_task_updated, ui_task_delta)  # ui event
            await con.commit()
            if len(tasks_to_wait) > 0:
                await asyncio.wait(tasks_to_wait)
        self.wake()
        self.poke_task_processor()

    async def _save_external_logs(self, invocation_id, stdout, stderr):
        logbasedir = self.__external_log_location / 'invocations' / f'{invocation_id}'
        try:
            if not logbasedir.exists():
                logbasedir.mkdir(exist_ok=True)
            async with aiofiles.open(logbasedir / 'stdout.log', 'w') as fstdout, \
                    aiofiles.open(logbasedir / 'stderr.log', 'w') as fstderr:
                await asyncio.gather(fstdout.write(stdout),
                                     fstderr.write(stderr))
        except OSError:
            self.__logger.exception('error happened saving external logs! Ignoring this error')

    #
    # worker reports canceled task
    async def task_cancel_reported(self, task: InvocationJob, stdout: str, stderr: str):
        async with self.__invocation_reporting_lock,\
                   self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            self.__logger.debug('task cancelled reported %s', repr(task))
            # sanity check
            async with con.execute('SELECT "state" FROM invocations WHERE "id" = ?', (task.invocation_id(),)) as cur:
                invoc = await cur.fetchone()
                if invoc is None:
                    self.__logger.error('reported task has non existing invocation id %d' % task.invocation_id())
                    return
                if invoc['state'] != InvocationState.IN_PROGRESS.value:
                    self.__logger.warning('reported task for a finished invocation. assuming that worker failed to cancel task previously and ignoring invocation results.')
                    return
            await con.execute('UPDATE invocations SET "state" = ?, "runtime" = ? WHERE "id" = ?',
                              (InvocationState.FINISHED.value, task.running_time(), task.invocation_id()))
            async with con.execute('SELECT * FROM invocations WHERE "id" = ?', (task.invocation_id(),)) as incur:
                invocation = await incur.fetchone()
            assert invocation is not None
            if task.invocation_id() in self.data_access.mem_cache_invocations:
                del self.data_access.mem_cache_invocations[task.invocation_id()]

            await con.execute('UPDATE workers SET "state" = ? WHERE "id" = ?',
                              (WorkerState.IDLE.value, invocation['worker_id']))
            await self._update_worker_resouce_usage(invocation['worker_id'], connection=con)  # remove resource usage info
            tasks_to_wait = []
            if not self.__use_external_log:
                await con.execute('UPDATE invocations SET "stdout" = ?, "stderr" = ? WHERE "id" = ?',
                                  (stdout, stderr, task.invocation_id()))
            else:
                await con.execute('UPDATE invocations SET "log_external" = 1, "stdout" = null, "stderr" = null WHERE "id" = ?',
                                  (task.invocation_id(),))
                tasks_to_wait.append(asyncio.create_task(self._save_external_logs(task.invocation_id(), stdout, stderr)))
            await con.execute('UPDATE tasks SET "state" = ? WHERE "id" = ?',
                              (TaskState.WAITING.value, invocation['task_id']))
            con.add_after_commit_callback(self.ui_state_access.scheduler_reports_task_updated, TaskDelta(invocation['task_id'], state=TaskState.WAITING))  # ui event
            await con.commit()
            if len(tasks_to_wait) > 0:
                await asyncio.wait(tasks_to_wait)
        self.__logger.debug(f'cancelling task done {repr(task)}')
        self.wake()
        self.poke_task_processor()

    #
    # add new worker to db
    async def add_worker(self, addr: str, worker_type: WorkerType, worker_resources: WorkerResources, assume_active=True):  # TODO: all resource should also go here
        self.__logger.debug(f'worker reported added: {addr}')
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            await con.execute('BEGIN IMMEDIATE')  # important to have locked DB during all this state change
            # logic for now:
            #  - search for same last_address, same hwid
            #  - if no - search for first entry (OFF or UNKNOWN) with same hwid, ignore address
            #    - in this case also delete addr from DB if exists
            async with con.execute('SELECT "id", state FROM "workers" WHERE "last_address" == ? AND hwid == ?', (addr, worker_resources.hwid)) as worcur:
                worker_row = await worcur.fetchone()
            if worker_row is None:
                # first ensure that there is no entry with the same address
                await con.execute('UPDATE "workers" SET "last_address" = ? WHERE "last_address" == ?', (None, addr))
                async with con.execute('SELECT "id", state FROM "workers" WHERE hwid == ? AND '
                                       '(state == ? OR state == ?)', (worker_resources.hwid,
                                                                      WorkerState.OFF.value, WorkerState.UNKNOWN.value)) as worcur:
                    worker_row = await worcur.fetchone()
            if assume_active:
                ping_state = WorkerPingState.WORKING.value
                state = WorkerState.IDLE.value
            else:
                ping_state = WorkerPingState.OFF.value
                state = WorkerState.OFF.value

            tstamp = int(time.time())
            if worker_row is not None:
                if worker_row['state'] == WorkerState.INVOKING.value:  # so we are in the middle of sumbission
                    state = WorkerState.INVOKING.value  # then we preserve INVOKING state
                await self.reset_invocations_for_worker(worker_row['id'], con=con, also_update_resources=False)  # we update later
                await con.execute('UPDATE "workers" SET '
                                  'hwid=?, '
                                  'last_seen=?, ping_state=?, state=?, worker_type=?, '
                                  'last_address=?  '
                                  'WHERE "id"=?',
                                  (worker_resources.hwid,
                                   tstamp, ping_state, state, worker_type.value,
                                   addr,
                                   worker_row['id']))
                # async with con.execute('SELECT "id" FROM "workers" WHERE last_address=?', (addr,)) as worcur:
                #     worker_id = (await worcur.fetchone())['id']
                worker_id = worker_row['id']
                self.data_access.mem_cache_workers_state[worker_id].update({'last_seen': tstamp,
                                                                        'last_checked': tstamp,
                                                                        'ping_state': ping_state,
                                                                        'worker_id': worker_id})
                # await con.execute('UPDATE tmpdb.tmp_workers_states SET '
                #                   'last_seen=?, ping_state=? '
                #                   'WHERE worker_id=?',
                #                   (tstamp, ping_state, worker_id))
            else:
                async with con.execute('INSERT INTO "workers" '
                                       '(hwid, '
                                       'last_address, last_seen, ping_state, state, worker_type) '
                                       'VALUES '
                                       '(?, ?, ?, ?, ?, ?)',
                                       (worker_resources.hwid, addr, tstamp, ping_state, state, worker_type.value)) as insworcur:
                    worker_id = insworcur.lastrowid
                self.data_access.mem_cache_workers_state[worker_id] = {'last_seen': tstamp,
                                                               'last_checked': tstamp,
                                                               'ping_state': ping_state,
                                                               'worker_id': worker_id}
                # await con.execute('INSERT INTO tmpdb.tmp_workers_states '
                #                   '(worker_id, last_seen, ping_state) '
                #                   'VALUES '
                #                   '(?, ?, ?)',
                #                   (worker_id, tstamp, ping_state))

            await con.execute('INSERT INTO resources '
                              '(hwid, cpu_count, total_cpu_count, '
                              'cpu_mem, total_cpu_mem, '
                              'gpu_count, total_gpu_count, '
                              'gpu_mem, total_gpu_mem) '
                              'VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?) '
                              'ON CONFLICT(hwid) DO UPDATE SET '
                              'cpu_count=excluded.cpu_count, total_cpu_count=excluded.total_cpu_count, '
                              'cpu_mem=excluded.cpu_mem, total_cpu_mem=excluded.total_cpu_mem, '
                              'gpu_count=excluded.gpu_count, total_gpu_count=excluded.total_gpu_count, '
                              'gpu_mem=excluded.gpu_mem, total_gpu_mem=excluded.total_gpu_mem',
                              (worker_resources.hwid,
                               worker_resources.cpu_count,
                               worker_resources.total_cpu_count,
                               worker_resources.cpu_mem,
                               worker_resources.total_cpu_mem,
                               worker_resources.gpu_count,
                               worker_resources.total_gpu_count,
                               worker_resources.gpu_mem,
                               worker_resources.total_gpu_mem)
                               )
            await self._update_worker_resouce_usage(worker_id, hwid=worker_resources.hwid, connection=con)  # used resources are inited to none
            await con.commit()
        self.__logger.debug(f'finished worker reported added: {addr}')
        self.poke_task_processor()

    # TODO: add decorator that locks method from reentry or smth
    #  potentially a worker may report done while this works,
    #  or when scheduler picked worker and about to run this, which will lead to inconsistency warning
    #  NOTE!: so far it's always called from a STARTED transaction, so there should not be reentry possible
    #  But that is not enforced right now, easy to make mistake
    async def _update_worker_resouce_usage(self, worker_id: int, resources: Optional[dict] = None, *, hwid=None, connection: aiosqlite.Connection) -> bool:
        """
        updates resource information based on new worker resources usage
        as part of ongoing transaction
        Note: con SHOULD HAVE STARTED TRANSACTION, otherwise it might be not safe to call this

        :param worker_id:
        :param hwid: if hwid of worker_id is already known - provide it here to skip extra db query. but be SURE it's correct!
        :param connection: opened db connection. expected to have Row as row factory
        :return: if commit is needed on connection (if db set operation happened)
        """
        resource_fields = ('cpu_count', 'cpu_mem', 'gpu_count', 'gpu_mem')

        workers_resources = self.data_access.mem_cache_workers_resources
        if hwid is None:
            async with connection.execute('SELECT "hwid" FROM "workers" WHERE "id" == ?', (worker_id,)) as worcur:
                hwid = (await worcur.fetchone())['hwid']

        # calculate available resources NOT counting current worker_id
        async with connection.execute(f'SELECT '
                                      f'{", ".join(resource_fields)}, '
                                      f'{", ".join("total_"+x for x in resource_fields)} '
                                      f'FROM resources WHERE hwid == ?', (hwid,)) as rescur:
            available_res = dict(await rescur.fetchone())
        current_available = {k: v for k, v in available_res.items() if not k.startswith('total_')}
        available_res = {k[len('total_'):]: v for k, v in available_res.items() if k.startswith('total_')}

        for wid, res in workers_resources.items():
            if wid == worker_id:
                continue  # SKIP worker_id currently being set
            if res.get('hwid') != hwid:
                continue
            for field in resource_fields:
                if field not in res:
                    continue
                available_res[field] -= res[field]
        ##

        # now choose proper amount of resources to pick
        if resources is None:
            workers_resources[worker_id] = {'hwid': hwid}  # remove resource usage info
        else:
            workers_resources[worker_id] = {}
            for field in resource_fields:
                if field not in resources:
                    continue
                if available_res[field] < resources[field]:
                    raise NotEnoughResources(f'{field}: {resources[field]} out of {available_res[field]}')
                # so we take preferred amount of resources (or minimum if pref not set), but no more than available
                # if preferred is lower than min - it's ignored
                workers_resources[worker_id][field] = min(available_res[field],
                                                          max(resources.get(f'pref_{field}', resources[field]),
                                                              resources[field]))
                available_res[field] -= workers_resources[worker_id][field]

            workers_resources[worker_id]['hwid'] = hwid  # just to ensure it was not overriden

        self.__logger.debug(f'updating resources {hwid} with {available_res} against {current_available}')
        self.__logger.debug(workers_resources)

        if available_res == current_available:  # nothing needs to be updated
            return False

        await connection.execute(f'UPDATE resources SET {", ".join(f"{k}={v}" for k, v in available_res.items())} WHERE hwid == ?', (hwid,))
        return True

    #
    # worker reports it being stopped
    async def worker_stopped(self, addr: str):
        """

        :param addr:
        :return:
        """
        self.__logger.debug(f'worker reported stopped: {addr}')
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            await con.execute('BEGIN IMMEDIATE')
            async with con.execute('SELECT id, hwid from "workers" WHERE "last_address" = ?', (addr,)) as worcur:
                worker_row = await worcur.fetchone()
            if worker_row is None:
                self.__logger.warning(f'unregistered worker reported "stopped": {addr}, ignoring')
                await con.rollback()
                return
            wid = worker_row['id']
            hwid = worker_row['hwid']
            # print(wid)

            # we ensure there are no invocations running with this worker
            async with con.execute('SELECT "id", task_id FROM invocations WHERE worker_id = ? AND ("state" = ? OR "state" = ?)',
                                   (wid, InvocationState.IN_PROGRESS.value, InvocationState.INVOKING.value)) as invcur:
                invocations = await invcur.fetchall()

            await con.execute('UPDATE workers SET "state" = ? WHERE "id" = ?', (WorkerState.OFF.value, wid))
            await con.executemany('UPDATE invocations SET state = ? WHERE "id" = ?', ((InvocationState.FINISHED.value, x["id"]) for x in invocations))
            await con.executemany('UPDATE tasks SET state = ? WHERE "id" = ?', ((TaskState.WAITING.value, x["task_id"]) for x in invocations))
            await self._update_worker_resouce_usage(wid, hwid=hwid, connection=con)
            del self.data_access.mem_cache_workers_resources[wid]  # remove from cache
            if len(invocations) > 0:
                con.add_after_commit_callback(self.ui_state_access.scheduler_reports_tasks_updated, [TaskDelta(x["task_id"], state=TaskState.WAITING) for x in invocations])  # ui event
            await con.commit()
        self.__logger.debug(f'finished worker reported stopped: {addr}')

    #
    # protocol related commands
    #
    #
    # cancel invocation
    async def cancel_invocation(self, invocation_id: str):
        self.__logger.debug(f'canceling invocation {invocation_id}')
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            async with con.execute('SELECT * FROM "invocations" WHERE "id" = ?', (invocation_id,)) as cur:
                invoc = await cur.fetchone()
            if invoc is None or invoc['state'] != InvocationState.IN_PROGRESS.value:
                return
            async with con.execute('SELECT "last_address" FROM "workers" WHERE "id" = ?', (invoc['worker_id'],)) as cur:
                worker = await cur.fetchone()
        if worker is None:
            self.__logger.error('inconsistent worker ids? how?')
            return
        addr = AddressChain(worker['last_address'])

        # the logic is:
        # - we send the worker a signal to cancel invocation
        # - later worker sends task_cancel_reported, and we are happy
        # - but worker might be overloaded, broken or whatever and may never send it. and it can even finish task and send task_done_reported, witch we need to treat
        with WorkerControlClient.get_worker_control_client(addr, self.message_processor()) as client:  # type: WorkerControlClient
            await client.cancel_task()

        # oh no, we don't do that, we wait for worker to report task canceled.  await con.execute('UPDATE invocations SET "state" = ? WHERE "id" = ?', (InvocationState.FINISHED.value, invocation_id))

    #
    #
    async def cancel_invocation_for_task(self, task_id: int):
        self.__logger.debug(f'canceling invocation for task {task_id}')
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            async with con.execute('SELECT "id" FROM "invocations" WHERE "task_id" = ? AND state = ?', (task_id, InvocationState.IN_PROGRESS.value)) as cur:
                invoc = await cur.fetchone()
        if invoc is None:
            return
        return await self.cancel_invocation(invoc['id'])

    #
    #
    async def cancel_invocation_for_worker(self, worker_id: int):
        self.__logger.debug(f'canceling invocation for worker {worker_id}')
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            async with con.execute('SELECT "id" FROM "invocations" WHERE "worker_id" == ? AND state == ?', (worker_id, InvocationState.IN_PROGRESS.value)) as cur:
                invoc = await cur.fetchone()
        if invoc is None:
            return
        return await self.cancel_invocation(invoc['id'])

    #
    #
    async def force_set_node_task(self, task_id: int, node_id: int):
        self.__logger.debug(f'forcing task {task_id} to node {node_id}')
        try:
            async with self.data_access.data_connection() as con:
                con.row_factory = aiosqlite.Row
                await con.execute('PRAGMA FOREIGN_KEYS = on')
                await con.execute('UPDATE tasks SET "node_id" = ? WHERE "id" = ?', (node_id, task_id))
                con.add_after_commit_callback(self.ui_state_access.scheduler_reports_task_updated, TaskDelta(task_id, node_id=node_id))  # ui event
                await con.commit()
        except aiosqlite.IntegrityError:
            self.__logger.error('could not remove node connection because of database integrity check')
        else:
            self.wake()
            self.poke_task_processor()

    #
    # force change task state
    async def force_change_task_state(self, task_ids: Union[int, Iterable[int]], state: TaskState):
        """
        forces task into given state.
        obviously a task cannot be forced into certain states, like IN_PROGRESS, GENERATING, POST_GENERATING
        :param task_ids:
        :param state:
        :return:
        """
        if state in (TaskState.IN_PROGRESS, TaskState.GENERATING, TaskState.POST_GENERATING):
            self.__logger.error(f'cannot force task {task_ids} into state {state}')
            return
        if isinstance(task_ids, int):
            task_ids = [task_ids]
        query = 'UPDATE tasks SET "state" = %d WHERE "id" = ?' % state.value
        #print('beep')
        async with self.data_access.data_connection() as con:
            for task_id in task_ids:
                await con.execute('BEGIN IMMEDIATE')
                async with con.execute('SELECT "state" FROM tasks WHERE "id" = ?', (task_id,)) as cur:
                    cur_state = await cur.fetchone()
                    if cur_state is None:
                        await con.rollback()
                        continue
                    cur_state = TaskState(cur_state[0])
                if cur_state in (TaskState.IN_PROGRESS, TaskState.GENERATING, TaskState.POST_GENERATING):
                    self.__logger.warning(f'forcing task out of state {cur_state} is not currently implemented')
                    await con.rollback()
                    continue

                await con.execute(query, (task_id,))
                #await con.executemany(query, ((x,) for x in task_ids))
                con.add_after_commit_callback(self.ui_state_access.scheduler_reports_task_updated, TaskDelta(task_id, state=state))  # ui event
                await con.commit()  # TODO: this can be optimized into a single transaction
        #print('boop')
        self.wake()
        self.poke_task_processor()

    #
    # change task's paused state
    async def set_task_paused(self, task_ids_or_group: Union[int, Iterable[int], str], paused: bool):
        if isinstance(task_ids_or_group, str):
            async with self.data_access.data_connection() as con:
                await con.execute('UPDATE tasks SET "paused" = ? WHERE "id" IN (SELECT "task_id" FROM task_groups WHERE "group" = ?)',
                                  (int(paused), task_ids_or_group))
                ui_task_ids = await self.ui_state_access._get_group_tasks(task_ids_or_group)  # ui event
                con.add_after_commit_callback(self.ui_state_access.scheduler_reports_tasks_updated, [TaskDelta(ui_task_id, paused=paused) for ui_task_id in ui_task_ids])  # ui event
                await con.commit()
            self.wake()
            self.poke_task_processor()
            return
        if isinstance(task_ids_or_group, int):
            task_ids_or_group = [task_ids_or_group]
        query = 'UPDATE tasks SET "paused" = %d WHERE "id" = ?' % int(paused)
        async with self.data_access.data_connection() as con:
            await con.executemany(query, ((x,) for x in task_ids_or_group))
            con.add_after_commit_callback(self.ui_state_access.scheduler_reports_tasks_updated, [TaskDelta(ui_task_id, paused=paused) for ui_task_id in task_ids_or_group])  # ui event
            await con.commit()
        self.wake()
        self.poke_task_processor()

    #
    # change task group archived state
    async def set_task_group_archived(self, task_group_name: str, state: TaskGroupArchivedState = TaskGroupArchivedState.ARCHIVED) -> None:
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            await con.execute('UPDATE task_group_attributes SET state=? WHERE "group"==?', (state.value, task_group_name))  # this triggers all task deadness | 2, so potentially it can be long, beware
            # task's dead field's 2nd bit is set, but we currently do not track it
            # so no event needed
            await con.commit()
            if state == TaskGroupArchivedState.NOT_ARCHIVED:
                self.poke_task_processor()  # unarchived, so kick task processor, just in case
                return
            # otherwise - it's archived
            # now all tasks belonging to that group should be set to dead|2
            # we need to make sure to cancel all running invocations for those tasks
            # at this point tasks are archived and won't be processed,
            # so we only expect concurrent changes due to already running _submitters and _awaiters,
            # like INVOKING->IN_PROGRESS
            async with con.execute('SELECT "id" FROM invocations '
                                   'INNER JOIN task_groups ON task_groups.task_id == invocations.task_id '
                                   'WHERE task_groups."group" == ? AND invocations.state == ?',
                                   (task_group_name, InvocationState.INVOKING.value)) as cur:
                invoking_invoc_ids = set(x['id'] for x in await cur.fetchall())
            async with con.execute('SELECT "id" FROM invocations '
                                   'INNER JOIN task_groups ON task_groups.task_id == invocations.task_id '
                                   'WHERE task_groups."group" == ? AND invocations.state == ?',
                                   (task_group_name, InvocationState.IN_PROGRESS.value)) as cur:
                active_invoc_ids = tuple(x['id'] for x in await cur.fetchall())
                # i sure use a lot of fetchall where it's much more natural to iterate cursor
                # that is because of a fear of db locking i got BEFORE switching to WAL, when iterating connection was randomly crashing other connections not taking timeout into account at all.

        # note at this point we might have some invoking_invocs_id, but at this point some of them
        # might already have been set to in-progress and even got into active_invoc_ids list

        # first - cancel all in-progress invocations
        for inv_id in active_invoc_ids:
            await self.cancel_invocation(inv_id)

        # now since we dont have the ability to safely cancel running _submitter task - we will just wait till
        # invoking invocations change state
        # sure it's a bit bruteforce
        # but a working solution for now
        if len(invoking_invoc_ids) == 0:
            return
        async with self.data_access.data_connection() as con:
            while len(invoking_invoc_ids) > 0:
                # TODO: this forever while doesn't seem right
                #  in average case it should basically never happen at all
                #  only in case of really bad buggy network connections an invocation can get stuck on INVOKING
                #  but there are natural timeouts in _submitter that will switch it from INVOKING eventually
                #  the only question is - do we want to just stay in this function until it's resolved? UI's client is single thread, so it will get stuck waiting
                con.row_factory = aiosqlite.Row
                async with con.execute('SELECT "id",state FROM invocations WHERE state!={} AND "id" IN ({})'.format(
                                            InvocationState.IN_PROGRESS.value,
                                            ','.join(str(x) for x in invoking_invoc_ids))) as cur:
                    changed_state_ones = await cur.fetchall()

                for oid, ostate in ((x['id'], x['state']) for x in changed_state_ones):
                    if ostate == InvocationState.IN_PROGRESS.value:
                        await self.cancel_invocation(oid)
                    assert oid in invoking_invoc_ids
                    invoking_invoc_ids.remove(oid)
                await asyncio.sleep(0.5)

    #
    # set task name
    async def set_task_name(self, task_id: int, new_name: str):
        async with self.data_access.data_connection() as con:
            await con.execute('UPDATE tasks SET "name" = ? WHERE "id" = ?', (new_name, task_id))
            con.add_after_commit_callback(self.ui_state_access.scheduler_reports_task_updated, TaskDelta(task_id, name=new_name))  # ui event
            await con.commit()

    #
    # set task groups
    async def set_task_groups(self, task_id: int, group_names: Iterable[str]):
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            await con.execute('BEGIN IMMEDIATE')
            async with con.execute('SELECT "group" FROM task_groups WHERE "task_id" = ?', (task_id,)) as cur:
                all_groups = set(x['group'] for x in await cur.fetchall())
            group_names = set(group_names)
            groups_to_set = group_names - all_groups
            groups_to_del = all_groups - group_names
            print(task_id, groups_to_set, groups_to_del, all_groups, group_names)

            for group_name in groups_to_set:
                await con.execute('INSERT INTO task_groups (task_id, "group") VALUES (?, ?)', (task_id, group_name))
                await con.execute('INSERT OR IGNORE INTO task_group_attributes ("group", "ctime") VALUES (?, ?)', (group_name, int(datetime.utcnow().timestamp())))
            for group_name in groups_to_del:
                await con.execute('DELETE FROM task_groups WHERE task_id = ? AND "group" = ?', (task_id, group_name))
            con.add_after_commit_callback(self.ui_state_access.scheduler_reports_tasks_removed_from_group, [task_id], groups_to_del)  # ui event
            con.add_after_commit_callback(self.ui_state_access.scheduler_reports_task_groups_changed, groups_to_set)  # ui event
            #
            # ui event
            if len(groups_to_set) > 0:
                async with con.execute(
                        'SELECT tasks.id, tasks.parent_id, tasks.children_count, tasks.active_children_count, tasks.state, tasks.state_details, tasks.paused, tasks.node_id, '
                        'tasks.node_input_name, tasks.node_output_name, tasks.name, tasks.split_level, tasks.work_data_invocation_attempt, '
                        'task_splits.origin_task_id, task_splits.split_id, invocations."id" as invoc_id '
                        'FROM "tasks" '
                        'LEFT JOIN "task_splits" ON tasks.id=task_splits.task_id '
                        'LEFT JOIN "invocations" ON tasks.id=invocations.task_id AND invocations.state = ? '
                        'WHERE tasks."id" == ?',
                        (InvocationState.IN_PROGRESS.value, task_id)) as cur:
                    task_row = await cur.fetchone()
                if task_row is not None:
                    progress = self.data_access.mem_cache_invocations.get(task_row['invoc_id'], {}).get('progress', None)
                    con.add_after_commit_callback(
                        self.ui_state_access.scheduler_reports_task_added,
                        TaskData(task_id, task_row['parent_id'], task_row['children_count'], task_row['active_children_count'], TaskState(task_row['state']),
                                 task_row['state_details'], bool(task_row['paused']), task_row['node_id'], task_row['node_input_name'], task_row['node_output_name'],
                                 task_row['name'], task_row['split_level'], task_row['work_data_invocation_attempt'], progress,
                                 task_row['origin_task_id'], task_row['split_id'], task_row['invoc_id'], group_names),
                        groups_to_set
                    )  # ui event
            con.add_after_commit_callback(self.ui_state_access.scheduler_reports_task_updated, TaskDelta(task_id, groups=group_names))  # ui event
            #
            #
            await con.commit()

    #
    # update task attributes
    async def update_task_attributes(self, task_id: int, attributes_to_update: dict, attributes_to_delete: set):
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            await con.execute('BEGIN IMMEDIATE')
            async with con.execute('SELECT "attributes" FROM tasks WHERE "id" = ?', (task_id,)) as cur:
                row = await cur.fetchone()
            if row is None:
                self.__logger.warning(f'update task attributes for {task_id} failed. task id not found.')
                await con.commit()
                return
            attributes = await asyncio.get_event_loop().run_in_executor(None, json.loads, row['attributes'])
            attributes.update(attributes_to_update)
            for name in attributes_to_delete:
                if name in attributes:
                    del attributes[name]
            await con.execute('UPDATE tasks SET "attributes" = ? WHERE "id" = ?', (await asyncio.get_event_loop().run_in_executor(None, json.dumps, attributes),
                                                                                   task_id))
            await con.commit()

    #
    # set environment resolver
    async def set_task_environment_resolver_arguments(self, task_id: int, env_res: Optional[EnvironmentResolverArguments]):
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            await con.execute('UPDATE tasks SET "environment_resolver_data" = ? WHERE "id" = ?',
                              (await env_res.serialize_async() if env_res is not None else None,
                               task_id))
            await con.commit()

    #
    # node stuff
    async def set_node_name(self, node_id: int, node_name: str) -> str:
        """
        rename node. node_name may undergo validation and change. final node name that was set is returned
        :param node_id: node id
        :param node_name: proposed node name
        :return: actual node name set
        """
        async with self.data_access.data_connection() as con:
            await con.execute('UPDATE "nodes" SET "name" = ? WHERE "id" = ?', (node_name, node_id))
            if node_id in self.__node_objects:
                self.__node_objects[node_id].set_name(node_name)
            await con.commit()
            self.ui_state_access.bump_graph_update_id()
        return node_name

    #
    # reset node's stored state
    async def wipe_node_state(self, node_id):
        async with self.data_access.data_connection() as con:
            await con.execute('UPDATE "nodes" SET node_object = NULL WHERE "id" = ?', (node_id,))
            if node_id in self.__node_objects:
                # TODO: this below may be not safe (at least not proven to be safe yet, but maybe). check
                del self.__node_objects[node_id]  # it's here to "protect" operation within db transaction. TODO: but a proper __node_object lock should be in place instead
            await con.commit()
            self.ui_state_access.bump_graph_update_id()  # not sure if needed - even number of inputs/outputs is not part of graph description
        self.wake()

    #
    # copy nodes
    async def duplicate_nodes(self, node_ids: Iterable[int]) -> Dict[int, int]:
        """
        copies given nodes, including connections between given nodes,
        and returns mapping from given node_ids to respective new copies

        :param node_ids:
        :return:
        """
        old_to_new = {}
        for nid in node_ids:
            node_obj = await self._get_node_object_by_id(nid)
            node_type, node_name = await self.get_node_type_and_name_by_id(nid)
            new_id = await self.add_node(node_type, f'{node_name} copy')
            new_node_obj = await self._get_node_object_by_id(new_id)
            node_obj.copy_ui_to(new_node_obj)
            old_to_new[nid] = new_id

        # now copy connections
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            node_ids_str = f'({",".join(str(x) for x in node_ids)})'
            async with con.execute(f'SELECT * FROM node_connections WHERE node_id_in IN {node_ids_str} AND node_id_out IN {node_ids_str}') as cur:
                all_cons = await cur.fetchall()
        for nodecon in all_cons:
            assert nodecon['node_id_in'] in old_to_new
            assert nodecon['node_id_out'] in old_to_new
            await self.add_node_connection(old_to_new[nodecon['node_id_out']], nodecon['out_name'], old_to_new[nodecon['node_id_in']], nodecon['in_name'])
        return old_to_new
        # TODO: NotImplementedError("recheck and needs testing")

    #
    #
    # node reports it's interface was changed. not sure why it exists
    async def node_reports_changes_needs_saving(self, node_id):
        assert node_id in self.__node_objects, 'this may be caused by race condition with node deletion'
        # TODO: introduce __node_objects lock? or otherwise secure access
        await self.save_node_to_database(node_id)

    #
    # save node to database.
    async def save_node_to_database(self, node_id):
        """
        save node with given node_id to database
        if node is not in our list of nodes - we assume it was not touched, not changed, so no saving needed

        :param node_id:
        :return:
        """
        # TODO: introduce __node_objects lock? or otherwise secure access
        #  why? this happens on ui_update, which can happen cuz of request from viewer.
        #  while node processing happens in a different thread, so this CAN happen at the same time with this
        #  AND THIS IS BAD! (potentially) if a node has changing internal state - this can save some inconsistent snapshot of node state!
        node_object = self.__node_objects[node_id]
        if node_object is None:
            self.__logger.error('node_object is None while')
            return
        async with self.data_access.data_connection() as con:
            await con.execute('UPDATE "nodes" SET node_object = ? WHERE "id" = ?',
                              (await node_object.serialize_async(), node_id))
            await con.commit()

    #
    # set worker groups
    async def set_worker_groups(self, worker_hwid: int, groups: List[str]):
        groups = set(groups)
        async with self.data_access.data_connection() as con:
            await con.execute('BEGIN IMMEDIATE')  # start transaction straight away
            async with con.execute('SELECT "group" FROM worker_groups WHERE worker_hwid == ?', (worker_hwid,)) as cur:
                existing_groups = set(x[0] for x in await cur.fetchall())
            to_delete = existing_groups - groups
            to_add = groups - existing_groups
            if len(to_delete):
                await con.execute(f'DELETE FROM worker_groups WHERE worker_hwid == ? AND "group" IN ({",".join(("?",)*len(to_delete))})', (worker_hwid, *to_delete))
            if len(to_add):
                await con.executemany(f'INSERT INTO worker_groups (worker_hwid, "group") VALUES (?, ?)',
                                      ((worker_hwid, x) for x in to_add))
            await con.commit()

    #
    # change node connection callback
    async def change_node_connection(self, node_connection_id: int, new_out_node_id: Optional[int], new_out_name: Optional[str],
                                     new_in_node_id: Optional[int], new_in_name: Optional[str]):
        parts = []
        vals = []
        if new_out_node_id is not None:
            parts.append('node_id_out = ?')
            vals.append(new_out_node_id)
        if new_out_name is not None:
            parts.append('out_name = ?')
            vals.append(new_out_name)
        if new_in_node_id is not None:
            parts.append('node_id_in = ?')
            vals.append(new_in_node_id)
        if new_in_name is not None:
            parts.append('in_name = ?')
            vals.append(new_in_name)
        if len(vals) == 0:  # nothing to do
            return
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            vals.append(node_connection_id)
            await con.execute(f'UPDATE node_connections SET {", ".join(parts)} WHERE "id" = ?', vals)
            await con.commit()
        self.wake()
        self.ui_state_access.bump_graph_update_id()

    #
    # add node connection callback
    async def add_node_connection(self, out_node_id: int, out_name: str, in_node_id: int, in_name: str) -> int:
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            async with con.execute('INSERT OR REPLACE INTO node_connections (node_id_out, out_name, node_id_in, in_name) VALUES (?,?,?,?)',  # INSERT OR REPLACE here (and not OR ABORT or smth) to ensure lastrowid is set
                                   (out_node_id, out_name, in_node_id, in_name)) as cur:
                ret = cur.lastrowid
            await con.commit()
            self.wake()
            self.ui_state_access.bump_graph_update_id()
            return ret

    #
    # remove node connection callback
    async def remove_node_connection(self, node_connection_id: int):
        try:
            async with self.data_access.data_connection() as con:
                con.row_factory = aiosqlite.Row
                await con.execute('PRAGMA FOREIGN_KEYS = on')
                await con.execute('DELETE FROM node_connections WHERE "id" = ?', (node_connection_id,))
                await con.commit()
                self.ui_state_access.bump_graph_update_id()
        except aiosqlite.IntegrityError as e:
            self.__logger.error('could not remove node connection because of database integrity check')

    #
    # add node
    async def add_node(self, node_type: str, node_name: str) -> int:
        if node_type not in pluginloader.plugins:
            raise RuntimeError('unknown node type')
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            async with con.execute('INSERT INTO "nodes" ("type", "name") VALUES (?,?)',
                                   (node_type, node_name)) as cur:
                ret = cur.lastrowid
            await con.commit()
            self.ui_state_access.bump_graph_update_id()
            return ret

    async def remove_node(self, node_id: int) -> bool:
        try:
            async with self.data_access.data_connection() as con:
                con.row_factory = aiosqlite.Row
                await con.execute('PRAGMA FOREIGN_KEYS = on')
                await con.execute('DELETE FROM "nodes" WHERE "id" = ?', (node_id,))
                await con.commit()
                self.ui_state_access.bump_graph_update_id()
        except aiosqlite.IntegrityError as e:
            self.__logger.error('could not remove node connection because of database integrity check')
            return False
        return True

    #
    # query connections
    async def get_node_input_connections(self, node_id: int, input_name: Optional[str] = None):
        return await self.get_node_connections(node_id, True, input_name)

    async def get_node_output_connections(self, node_id: int, output_name: Optional[str] = None):
        return await self.get_node_connections(node_id, False, output_name)

    async def get_node_connections(self, node_id: int, query_input: bool = True, name: Optional[str] = None):
        if query_input:
            nodecol = 'node_id_in'
            namecol = 'in_name'
        else:
            nodecol = 'node_id_out'
            namecol = 'out_name'
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            if name is None:
                async with con.execute('SELECT * FROM node_connections WHERE "%s" = ?' % (nodecol,),
                                       (node_id,)) as cur:
                    return [dict(x) for x in await cur.fetchall()]
            else:
                async with con.execute('SELECT * FROM node_connections WHERE "%s" = ? AND "%s" = ?' % (nodecol, namecol),
                                       (node_id, name)) as cur:
                    return [dict(x) for x in await cur.fetchall()]

    #
    # spawning new task callback
    @alocking()
    async def spawn_tasks(self, newtasks: Union[Iterable[TaskSpawn], TaskSpawn], con: Optional[aiosqlite_overlay.ConnectionWithCallbacks] = None) -> Union[Tuple[SpawnStatus, Optional[int]], Tuple[Tuple[SpawnStatus, Optional[int]], ...]]:
        """

        :param newtasks:
        :param con:
        :return:
        """

        async def _inner_shit() -> Tuple[Tuple[SpawnStatus, Optional[int]], ...]:
            result = []
            new_tasks = []
            current_timestamp = int(datetime.utcnow().timestamp())
            for newtask in newtasks:
                if newtask.source_invocation_id() is not None:
                    async with con.execute('SELECT node_id, task_id FROM invocations WHERE "id" = ?',
                                           (newtask.source_invocation_id(),)) as incur:
                        invocrow = await incur.fetchone()
                        assert invocrow is not None
                        node_id: int = invocrow['node_id']
                        parent_task_id: int = invocrow['task_id']
                elif newtask.forced_node_task_id() is not None:
                    node_id, parent_task_id = newtask.forced_node_task_id()
                else:
                    self.__logger.error('ERROR CREATING SPAWN TASK: Malformed source')
                    result.append((SpawnStatus.FAILED, None))
                    continue

                async with con.execute('INSERT INTO tasks ("name", "attributes", "parent_id", "state", "node_id", "node_output_name", "environment_resolver_data") VALUES (?, ?, ?, ?, ?, ?, ?)',
                                       (newtask.name(), json.dumps(newtask._attributes()), parent_task_id,  # TODO: run dumps in executor
                                        TaskState.SPAWNED.value if newtask.create_as_spawned() else TaskState.WAITING.value,
                                        node_id, newtask.node_output_name(),
                                        newtask.environment_arguments().serialize() if newtask.environment_arguments() is not None else None)) as newcur:
                    new_id = newcur.lastrowid

                all_groups = set()
                if parent_task_id is not None:  # inherit all parent's groups
                    # check and inherit parent's environment wrapper arguments
                    if newtask.environment_arguments() is None:
                        await con.execute('UPDATE tasks SET environment_resolver_data = (SELECT environment_resolver_data FROM tasks WHERE "id" == ?) WHERE "id" == ?',
                                          (parent_task_id, new_id))

                    # inc children count happens in db trigger
                    # inherit groups
                    async with con.execute('SELECT "group" FROM task_groups WHERE "task_id" = ?', (parent_task_id,)) as gcur:
                        groups = [x['group'] for x in await gcur.fetchall()]
                    all_groups.update(groups)
                    if len(groups) > 0:
                        con.add_after_commit_callback(self.ui_state_access.scheduler_reports_task_groups_changed, groups)  # ui event
                        await con.executemany('INSERT INTO task_groups ("task_id", "group") VALUES (?, ?)',
                                              zip(itertools.repeat(new_id, len(groups)), groups))
                else:  # parent_task_id is None
                    # in this case we create a default group for the task.
                    # task should not be left without groups at all - otherwise it will be impossible to find in UI
                    new_group = '{name}#{id:d}'.format(name=newtask.name(), id=new_id)
                    all_groups.add(new_group)
                    await con.execute('INSERT INTO task_groups ("task_id", "group") VALUES (?, ?)',
                                      (new_id, new_group))
                    await con.execute('INSERT OR REPLACE INTO task_group_attributes ("group", "ctime") VALUES (?, ?)',
                                      (new_group, current_timestamp))
                    if newtask.default_priority() is not None:
                        await con.execute('UPDATE task_group_attributes SET "priority" = ? WHERE "group" = ?',
                                          (newtask.default_priority(), new_group))
                    con.add_after_commit_callback(self.ui_state_access.scheduler_reports_task_groups_changed, (new_group,))  # ui event
                    #
                if newtask.extra_group_names():
                    groups = newtask.extra_group_names()
                    all_groups.update(groups)
                    await con.executemany('INSERT INTO task_groups ("task_id", "group") VALUES (?, ?)',
                                          zip(itertools.repeat(new_id, len(groups)), groups))
                    con.add_after_commit_callback(self.ui_state_access.scheduler_reports_task_groups_changed, groups)  # ui event
                    for group in groups:
                        async with con.execute('SELECT "group" FROM task_group_attributes WHERE "group" == ?', (group,)) as gcur:
                            need_create = await gcur.fetchone() is None
                        if not need_create:
                            continue
                        await con.execute('INSERT INTO task_group_attributes ("group", "ctime") VALUES (?, ?)',
                                          (group, current_timestamp))
                        # TODO: task_groups.group should be a foreign key to task_group_attributes.group
                        #  but then we need to insert those guys in correct order (first in attributes table, then groups)
                        #  then smth like FOREIGN KEY("group") REFERENCES "task_group_attributes"("group") ON UPDATE CASCADE ON DELETE CASCADE
                result.append((SpawnStatus.SUCCEEDED, new_id))
                new_tasks.append(TaskData(new_id, parent_task_id, 0, 0,
                                          TaskState.SPAWNED if newtask.create_as_spawned() else TaskState.WAITING, '',
                                          False, node_id, 'main', newtask.node_output_name(), newtask.name(), 0, 0, None, None, None, None,
                                          all_groups))

            # callbacks for ui events
            con.add_after_commit_callback(self.ui_state_access.scheduler_reports_tasks_added, new_tasks)
            return tuple(result)

        return_single = False
        if isinstance(newtasks, TaskSpawn):
            newtasks = (newtasks,)
            return_single = True
        if con is not None:
            stuff = await _inner_shit()
        else:
            async with self.data_access.data_connection() as con:
                con.row_factory = aiosqlite.Row
                stuff = await _inner_shit()
                await con.commit()
        self.wake()
        self.poke_task_processor()
        return stuff[0] if return_single else stuff

    #
    async def node_name_to_id(self, name: str) -> List[int]:
        """
        get the list of node ids that have specified name
        :param name:
        :return:
        """
        async with self.data_access.data_connection() as con:
            async with con.execute('SELECT "id" FROM "nodes" WHERE "name" = ?', (name,)) as cur:
                return list(x[0] for x in await cur.fetchall())

    #
    async def get_invocation_metadata(self, task_id: int) -> Dict[int, List[IncompleteInvocationLogData]]:
        """
        get task's log metadata - meaning which nodes it ran on and how
        :param task_id:
        :return: dict[node_id -> list[IncompleteInvocationLogData]]
        """
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            logs = {}
            self.__logger.debug(f'fetching log metadata for {task_id}')
            async with con.execute('SELECT "id", node_id, runtime, worker_id from "invocations" WHERE "task_id" = ?',
                                   (task_id, )) as cur:
                async for entry in cur:
                    node_id = entry['node_id']
                    logs.setdefault(node_id, []).append(IncompleteInvocationLogData(
                        entry['id'],
                        entry['worker_id'],
                        entry['runtime']
                    ))
            return logs

    async def get_log(self, invocation_id: int) -> Optional[InvocationLogData]:
        """
        get logs for given task, node and invocation ids

        returns a dict of node_id

        :param invocation_id:
        :return:
        """
        async with self.data_access.data_connection() as con:
            con.row_factory = aiosqlite.Row
            self.__logger.debug(f"fetching for {invocation_id}")
            async with con.execute('SELECT "id", task_id, worker_id, node_id, state, return_code, log_external, runtime, stdout, stderr '
                                   'FROM "invocations" WHERE "id" = ?',
                                   (invocation_id,)) as cur:
                rawentry = await cur.fetchone()  # should be exactly 1 or 0
            if rawentry is None:
                return None

            entry: InvocationLogData = InvocationLogData(rawentry['id'],
                                                         rawentry['worker_id'],
                                                         rawentry['runtime'],
                                                         rawentry['task_id'],
                                                         rawentry['node_id'],
                                                         InvocationState(rawentry['state']),
                                                         rawentry['return_code'],
                                                         rawentry['stdout'] or '',
                                                         rawentry['stderr'] or '')
            if entry.invocation_state == InvocationState.IN_PROGRESS:
                async with con.execute('SELECT last_address FROM workers WHERE "id" = ?', (entry.worker_id,)) as worcur:
                    workrow = await worcur.fetchone()
                if workrow is None:
                    self.__logger.error('Worker not found during log fetch! this is not supposed to happen! Database inconsistent?')
                else:
                    try:
                        with WorkerControlClient.get_worker_control_client(AddressChain(workrow['last_address']), self.message_processor()) as client:  # type: WorkerControlClient
                            stdout, stderr = await client.get_log(invocation_id)
                        if not self.__use_external_log:
                            await con.execute('UPDATE "invocations" SET stdout = ?, stderr = ? WHERE "id" = ?',  # TODO: is this really needed? if it's never really read
                                              (stdout, stderr, invocation_id))
                            await con.commit()
                        # TODO: maybe add else case? save partial log to file?
                    except ConnectionError:
                        self.__logger.warning('could not connect to worker to get freshest logs')
                    else:
                        entry.stdout = stdout
                        entry.stderr = stderr

            elif entry.invocation_state == InvocationState.FINISHED and rawentry['log_external'] == 1:
                logbasedir = self.__external_log_location / 'invocations' / f'{invocation_id}'
                stdout_path = logbasedir / 'stdout.log'
                stderr_path = logbasedir / 'stderr.log'
                try:
                    if stdout_path.exists():
                        async with aiofiles.open(stdout_path, 'r') as fstdout:
                            entry.stdout = await fstdout.read()
                except IOError:
                    self.__logger.exception(f'could not read external stdout log for {invocation_id}')
                try:
                    if stderr_path.exists():
                        async with aiofiles.open(stderr_path, 'r') as fstderr:
                            entry.stderr = await fstderr.read()
                except IOError:
                    self.__logger.exception(f'could not read external stdout log for {invocation_id}')

        return entry

    def server_address(self) -> str:
        return self.__server_address

    def server_message_address(self) -> AddressChain:
        if self.__message_processor is None:
            raise RuntimeError('cannot get listening address of a non started server')
        return self.message_processor().listening_address()



import asyncio
import socket
from string import ascii_letters
import random
import struct

from . import logging
from .nethelpers import get_localhost
from .defaults import broadcast_port as default_broadcast_port

from . import os_based_cheats

from typing import Tuple, Union, Optional, Callable, Coroutine, Any

Address = Tuple[str, int]

# base taken from https://gist.github.com/yluthu/4f785d4546057b49b56c

_magic = b'=\xe2\x88\x88\xe2\xad\x95\xe2\x88\x8b='


async def create_broadcaster(identifier: str, information, *, broad_port: int = default_broadcast_port(), ip='0.0.0.0', broadcasts_count: Optional[int] = None, broadcast_interval: int = 10):
    loop = asyncio.get_event_loop()
    return await loop.create_datagram_endpoint(
        lambda: BroadcastProtocol(identifier, information, (ip, broad_port), broadcasts_count, broadcast_interval, loop=loop),
        family=socket.AF_INET, reuse_port=True, allow_broadcast=True)


async def await_broadcast(identifier: str, broad_port: int = default_broadcast_port(), listen_address: str = '0.0.0.0') -> str:
    loop = asyncio.get_event_loop()
    protocol: BroadcastReceiveProtocol
    _, protocol = await loop.create_datagram_endpoint(
        lambda: BroadcastReceiveProtocol(identifier, loop=loop),
        local_addr=(listen_address, broad_port), family=socket.AF_INET, reuse_port=True, allow_broadcast=True)
    message = await protocol.till_done()

    # rebroadcast to localhost
    _, protocol = await create_broadcaster(identifier, message, broad_port=broad_port, ip=get_localhost(), broadcasts_count=1)  # TODO: bet with my future self - i will regret hardcoding ipv4 localhost like that
    await protocol.till_done()
    return message


class BroadcastListener:  # TODO: This was never tested
    def __init__(self, identifier: str, callback: Callable[[str], Coroutine], broad_port: int = default_broadcast_port(), listen_address: str = '0.0.0.0'):
        self.__identifier = identifier
        self.__callback = callback
        self.__addr = (listen_address, broad_port)
        self.__proto: Optional[BroadcastReceiveProtocol] = None
        self.__trans = None

    async def start(self):
        loop = asyncio.get_event_loop()
        self.__trans, self.__proto = await loop.create_datagram_endpoint(
            lambda: BroadcastReceiveProtocol(self.__identifier, loop=loop, single_shot=False, callback=self.__callback),
            local_addr=self.__addr, family=socket.AF_INET, reuse_port=True, allow_broadcast=True)

    def stop(self):
        self.__proto.close()

    async def wait_till_stopped(self):
        await self.__proto.till_done()


class BroadcastProtocol(asyncio.DatagramProtocol):
    def __init__(self, identifier: str, information: str, target: Address, count=None, broadcast_interval=10, *, loop: asyncio.AbstractEventLoop = None):
        self.__logger = logging.get_logger('broadcast')
        self.__target = target
        self.__loop = asyncio.get_event_loop() if loop is None else loop
        self.__transport = None
        self.__identifier = identifier
        self.__interval = broadcast_interval
        self.__broadcast_count = count
        self.__information = information
        self.__closed_future = self.__loop.create_future()
        self.__broadcast_task = None
        # TODO: raise on dgram size too big

    def connection_made(self, transport: asyncio.transports.DatagramTransport):
        self.__logger.info(f'started broadcasting')
        self.__transport = transport
        sock = transport.get_extra_info("socket")  # type: socket.socket
        #sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        #sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        self.__broadcast_task = self.__loop.create_task(self.broadcast())

    def datagram_received(self, data: Union[bytes, str], addr: Address):
        self.__logger.debug(f'data received: {len(data)}, {data[:64]}, {addr}')

    async def broadcast(self):
        self.__logger.info(f'starting broadcast for {self.__identifier} on {self.__target}')
        while True:
            self.__logger.debug(f'sending broadcast for {self.__identifier} on {self.__target}')
            bitent = self.__identifier.encode('UTF-8')
            binfo = self.__information.encode('UTF-8')
            data: bytes = b''.join((_magic, struct.pack('>I', len(bitent)), struct.pack('>I', len(binfo)), bitent, binfo))
            self.__transport.sendto(data, self.__target)
            if self.__broadcast_count is not None:
                self.__broadcast_count -= 1
                if self.__broadcast_count <= 0:
                    break
            await asyncio.sleep(self.__interval)
        self.__transport.close()
        self.__closed_future.set_result(True)

    async def till_done(self):
        return await self.__closed_future


class BroadcastReceiveProtocol(asyncio.DatagramProtocol):
    def __init__(self, identifier: str, loop: asyncio.AbstractEventLoop = None, single_shot: bool = True, callback: Optional[Callable[[str], Any]] = None):
        self.__logger = logging.get_logger('broadcast_catcher')
        self.__loop = asyncio.get_event_loop() if loop is None else loop
        self.__single_shot = single_shot
        self.__callback = callback
        self.__transport = None
        self.__identifier = identifier
        self.__last_message = None
        self.__closed_future = self.__loop.create_future()

    def connection_made(self, transport: asyncio.transports.DatagramTransport):
        self.__transport = transport
        sock = transport.get_extra_info("socket")  # type: socket.socket
        self.__logger.info(f'started listening on {sock.getsockname()}')
        #sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        #sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 255)
        #sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)

    def datagram_received(self, data: Union[bytes, str], addr: Address):
        self.__logger.debug(f'data received: {len(data)}, {data[:64]}, {addr}')
        if not data.startswith(_magic):
            return
        data = data[len(_magic):]
        identlen, bodylen = struct.unpack('>II', data[:8])
        identifier = data[8:8+identlen].decode('UTF-8')
        message = data[8+identlen:8+identlen+bodylen].decode('UTF-8')
        self.__logger.info(f'broadcast received from {identifier} at {addr}')
        self.__logger.debug(f'received: {identifier}, {message}')
        self.__last_message = message
        if self.__callback:
            self.__callback(message)
        if self.__single_shot:
            self.close()

    def close(self):
        self.__transport.close()
        self.__closed_future.set_result(self.__last_message)

    async def till_done(self):
        return await self.__closed_future


if __name__ == '__main__':

    async def message_receiver():
        result = await await_broadcast('teststuff')
        print('got result', result)

    asyncio.get_event_loop().create_task(create_broadcaster('teststuff', f'{get_localhost()}:6669'))
    asyncio.get_event_loop().create_task(message_receiver())
    asyncio.get_event_loop().run_forever()



import time
from io import BufferedIOBase
import struct
from dataclasses import dataclass, field
from .buffered_connection import BufferedReader
from .ui_protocol_data import TaskData, TaskDelta, TaskBatchData, UiData
from .buffer_serializable import IBufferSerializable
from .enums import UIEventType

from typing import ClassVar, Dict, Iterable, List, Tuple, Type, Union


@dataclass
class SchedulerEvent(IBufferSerializable):
    # some OS (like windows) do not have proper time() resolution, so time_ns is used instead
    timestamp: float = field(default_factory=lambda: time.time_ns(), init=False)
    event_id: int
    event_type: UIEventType
    database_uid: int

    def __hash__(self):
        return hash((self.event_id, self.timestamp, self.event_type))


@dataclass
class SchedulerEventAutoId(SchedulerEvent):
    event_id: int = field(default=-1, init=False)


@dataclass
class TaskEvent(SchedulerEventAutoId):
    """
    this is a base class for events
    """
    __subclass_to_id: ClassVar[Dict[Type["TaskEvent"], int]] = {}
    __id_to_subclass: ClassVar[Dict[int, Type["TaskEvent"]]] = {}
    __next_subclass_id: ClassVar[int] = 0

    @classmethod
    def register_subclasses(cls, subclasses_list: Iterable[Type["TaskEvent"]]):
        for subclass in subclasses_list:
            sid = cls.__next_subclass_id
            cls.__next_subclass_id += 1
            cls.__subclass_to_id[subclass] = sid
            cls.__id_to_subclass[sid] = subclass

    def serialize(self, stream: BufferedIOBase):
        """
        we serialize only the part that we know
        """
        stream.write(struct.pack('>BQQQI', self.__subclass_to_id[type(self)], self.database_uid, self.event_id, self.timestamp, self.event_type.value))

    @classmethod
    def deserialize(cls, stream: BufferedReader) -> "TaskEvent":
        """
        unlike all other deserializations, here we will deserialize into a subclass
        """
        sid, dbuid, event_id, timestamp, event_type_raw = struct.unpack('>BQQQI', stream.readexactly(29))
        event_type = UIEventType(event_type_raw)
        base_event = TaskEvent(event_type, dbuid)
        base_event.event_id = event_id
        base_event.timestamp = timestamp

        subclass = cls.__id_to_subclass[sid]
        return subclass._deserialize_part(base_event, stream)

    @classmethod
    def _deserialize_part(cls, base_event: "TaskEvent", stream: BufferedReader) -> "TaskEvent":
        raise NotImplementedError()

    def tiny_repr(self) -> str:
        return f'<{type(self).__name__}:{self.event_id}.{self.event_type.name}'


@dataclass
class TaskFullState(TaskEvent):
    task_data: TaskBatchData
    event_type: UIEventType = field(default=UIEventType.FULL_STATE, init=False)

    def serialize(self, stream: BufferedIOBase):
        super().serialize(stream)
        self.task_data.serialize(stream)

    @classmethod
    def _deserialize_part(cls, base_event: TaskEvent, stream: BufferedReader) -> "TaskFullState":
        task_data = TaskBatchData.deserialize(stream)
        event = TaskFullState(base_event.database_uid, task_data)
        assert event.event_type == base_event.event_type == UIEventType.FULL_STATE
        event.timestamp = base_event.timestamp
        event.event_id = base_event.event_id
        return event

    def tiny_repr(self) -> str:
        return super().tiny_repr() + f':{self.task_data.tiny_repr()}>'


@dataclass
class TasksChanged(TaskEvent):
    task_deltas: List[TaskDelta]
    event_type: UIEventType = field(default=UIEventType.UPDATE, init=False)

    def serialize(self, stream: BufferedIOBase):
        super().serialize(stream)
        stream.write(struct.pack('>Q', len(self.task_deltas)))
        for task_delta in self.task_deltas:
            task_delta.serialize(stream)

    @classmethod
    def _deserialize_part(cls, base_event: "TaskEvent", stream: BufferedReader) -> "TaskEvent":
        task_deltas = []
        num_deltas, = struct.unpack('>Q', stream.readexactly(8))
        for _ in range(num_deltas):
            task_deltas.append(TaskDelta.deserialize(stream))
        event = TasksChanged(base_event.database_uid, task_deltas)
        assert event.event_type == base_event.event_type == UIEventType.UPDATE
        event.timestamp = base_event.timestamp
        event.event_id = base_event.event_id
        return event

    def tiny_repr(self) -> str:
        return super().tiny_repr() + f':[{",".join(str(x.id) for x in self.task_deltas)}]>'


@dataclass
class TasksUpdated(TaskEvent):
    task_data: TaskBatchData
    event_type: UIEventType = field(default=UIEventType.UPDATE, init=False)

    def serialize(self, stream: BufferedIOBase):
        super().serialize(stream)
        self.task_data.serialize(stream)

    @classmethod
    def _deserialize_part(cls, base_event: TaskEvent, stream: BufferedReader) -> "TasksUpdated":
        task_data = TaskBatchData.deserialize(stream)
        event = TasksUpdated(base_event.database_uid, task_data)
        assert event.event_type == base_event.event_type == UIEventType.UPDATE
        event.timestamp = base_event.timestamp
        event.event_id = base_event.event_id
        return event

    def tiny_repr(self) -> str:
        return super().tiny_repr() + f':{self.task_data.tiny_repr()}>'

@dataclass
class TasksRemoved(TaskEvent):
    task_ids: Tuple[int, ...]  # task data, or task_id depending on UIEventType
    event_type: UIEventType = field(default=UIEventType.DELETE, init=False)

    def serialize(self, stream: BufferedIOBase):
        super().serialize(stream)
        stream.write(struct.pack('>Q', len(self.task_ids)))
        for task_id in self.task_ids:
            stream.write(struct.pack('>Q', task_id))

    @classmethod
    def _deserialize_part(cls, base_event: TaskEvent, stream: BufferedReader) -> "TasksRemoved":
        task_count, = struct.unpack('>Q', stream.readexactly(8))
        task_ids = struct.unpack('>' + 'Q' * task_count, stream.readexactly(8 * task_count))
        event = TasksRemoved(base_event.database_uid, task_ids)
        assert event.event_type == base_event.event_type == UIEventType.DELETE
        event.timestamp = base_event.timestamp
        event.event_id = base_event.event_id
        return event

    def tiny_repr(self) -> str:
        return super().tiny_repr() + f':{self.task_ids}>'


TaskEvent.register_subclasses([TaskFullState, TasksChanged, TasksUpdated, TasksRemoved])



from lifeblood.basenode import ProcessingResult, ProcessingContext, ProcessingError
from lifeblood.invocationjob import InvocationJob
from lifeblood.enums import NodeParameterType
from lifeblood_alicevision_modules.base_node import AlicevisionBaseNode
from pathlib import Path
from typing import Iterable


def node_class():
    return AlicevisionImageMatching


class AlicevisionImageMatching(AlicevisionBaseNode):
    def __init__(self, name):
        super(AlicevisionImageMatching, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_parameter('input', 'Input Images Directory', NodeParameterType.STRING, "`task['av_camera_sfm']`")
            ui.add_parameter('featuresFolders', 'Features Dir', NodeParameterType.STRING, "`task['av_features_dir']`")  # TODO: this seem to be an array of dirs, so add support
            ui.add_parameter('output', 'Input Images Directory', NodeParameterType.STRING, "`config['global_scratch_location']`/alicevision/`task['uuid']`/ImageMatching/imageMatches.txt")
            ui.add_separator()
            ui.add_parameter('method', 'Method', NodeParameterType.STRING, 'VocabularyTree')\
                .add_menu((('Vocabulary Tree', 'VocabularyTree'),
                           ('Sequential', 'Sequential'),
                           ('Sequential + Vocabulary Tree', 'SequentialAndVocabularyTree'),
                           ('Exhaustive', 'Exhaustive'),
                           ('Frustum', 'Frustum'),
                           ('Frustum or Vocabulary Tree', 'FrustumOrVocabularyTree')))
            ui.add_parameter('tree', 'Vocabulary Tree', NodeParameterType.STRING, '../../aliceVision/share/aliceVision/vlfeat_K80L3.SIFT.tree')
            ui.add_parameter('minNbImages', 'minNbImages', NodeParameterType.INT, 200)
            ui.add_parameter('maxDescriptors', 'maxDescriptors', NodeParameterType.INT, 500)
            ui.add_parameter('nbMatches', 'nbMatches', NodeParameterType.INT, 50)
            ui.add_parameter('weights', 'weights', NodeParameterType.STRING, '')

    @classmethod
    def label(cls) -> str:
        return 'AV Image Matching'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'alicevision', 'image', 'matching'

    @classmethod
    def type_name(cls) -> str:
        return 'stock_alicevision_image_matching'

    @classmethod
    def description(cls) -> str:
        return 'this node type does not have a description'

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        tree_path = Path(context.param_value('tree'))
        output = Path(context.param_value('output').strip())
        if not output.is_absolute():
            raise ProcessingError('output path must be absolute and not empty')

        output.parent.mkdir(parents=True, exist_ok=True)  # TODO: move this to worker

        args = ['aliceVision_imageMatching', '--verboseLevel', 'info']
        args += ['--input', context.param_value('input')]
        args += ['--featuresFolder', context.param_value('featuresFolders')]
        args += ['--output', output]
        args += ['--method', context.param_value('method')]
        args += ['--tree', tree_path]
        args += ['--minNbImages', context.param_value('minNbImages')]
        args += ['--maxDescriptors', context.param_value('maxDescriptors')]
        args += ['--nbMatches', context.param_value('nbMatches')]
        args += ['--weights', context.param_value('weights')]

        job = InvocationJob([str(x) for x in args])
        res = ProcessingResult(job)
        res.set_attribute('av_image_matches', str(output))
        return res



#!/usr/bin/env python3

import os
import sys
import subprocess
from pathlib import Path


def main():
    cwd = Path(os.getcwd())
    mcdir = cwd/'matrixcommander'

    syspy = sys.executable

    if mcdir.exists():
        print('{mcdir} already exists! aborting install')
        return False

    print(f'installing matrix_commander venv into {mcdir}')
    mcdir.mkdir()

    # create venv
    res = subprocess.Popen([syspy, '-m', 'venv', 'venv'], cwd=mcdir).wait()
    if res != 0:
        print('error creating venv')
        return False
    venvpy = mcdir/'venv'/'bin'/'python'

    res = subprocess.Popen([venvpy, '-m', 'pip', 'install', 'matrix-commander~=6.0'], cwd=mcdir).wait()
    if res != 0:
        print('error installing matrix-commander into venv')
        return False

    mc_entry = mcdir/'venv'/'bin'/'matrix-commander'
    res = subprocess.Popen([venvpy, mc_entry, '--version'], cwd=mcdir).wait()
    if res == 0:
        print('success')
        return True
    else:
        print('something failed')
        return False


if __name__ == '__main__':
    sys.exit(0 if main() else 1)



from lifeblood.basenode import BaseNodeWithTaskRequirements
from lifeblood.enums import NodeParameterType, WorkerType
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment
from lifeblood.text import filter_by_pattern


import os
import shutil
import tempfile

from typing import Iterable


def node_class():
    return CopyHipFile


class CopyHipFile(BaseNodeWithTaskRequirements):
    @classmethod
    def label(cls) -> str:
        return 'hip file operations start'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'hip', 'houdini', 'copy', 'start'

    @classmethod
    def type_name(cls) -> str:
        return 'copy hip file'

    @classmethod
    def description(cls) -> str:
        return ''

    def __init__(self, name):
        super(CopyHipFile, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.5, 0.25, 0.125)
            ui.add_parameter('on workers', 'submit to workers', NodeParameterType.BOOL, False)
            ui.add_parameter('hip path', 'source hip path', NodeParameterType.STRING, "`task['hipfile']`")
            ui.add_parameter('dst hip path', 'dest hip path', NodeParameterType.STRING, "`config['global_scratch_location']`/`node.name`")
            ui.add_parameter('append unique hipname', 'append unique temporary hip name to destination', NodeParameterType.BOOL, True)
            with ui.parameters_on_same_line_block():
                mask_hip = ui.add_parameter('save orig', 'save original hip path', NodeParameterType.BOOL, False)
                ui.add_parameter('mask hip path attribute', 'attribute name', NodeParameterType.STRING, "hipfile_orig")\
                    .append_visibility_condition(mask_hip, '==', True)

                ui.parameter('worker type').set_value(WorkerType.SCHEDULER_HELPER.value)

    def process_task(self, context) -> ProcessingResult:
        src_path = context.param_value('hip path')
        dst_path = context.param_value('dst hip path')

        if context.param_value('on workers'):
            script = 'import os, shutil, tempfile\n' \
                     'import lifeblood_connection\n' \
                    f'dst_path = {repr(dst_path)}\n'
            if context.param_value('append unique hipname'):
                script += "_fd, dst_path = tempfile.mkstemp(suffix='.hip', dir=dst_path)\n" \
                          'os.close(_fd)\n'
            script += 'dst_dir = os.path.dirname(dst_path)\n' \
                      'if not os.path.exists(dst_dir):\n' \
                      '    os.makedirs(dst_dir, exist_ok=True)\n' \
                     f'shutil.copy2({repr(src_path)}, dst_path)\n' \
                      'attrs_to_set = {}\n' \
                      'attrs_to_set["hipfile"] = dst_path\n'
            if context.param_value('save orig'):
                script += f'attrs_to_set[{repr(context.param_value("mask hip path attribute"))}] = {repr(src_path)}\n'
            script += 'lifeblood_connection.set_attributes(attrs_to_set, blocking=True)\n'
            job = InvocationJob(['python', ':/work_to_do.py'])
            job.set_extra_file('work_to_do.py', script)
            return ProcessingResult(job=job)

        if context.param_value('append unique hipname'):
            fd, dst_path = tempfile.mkstemp(suffix='.hip', dir=dst_path)
            os.close(fd)
        dst_dir = os.path.dirname(dst_path)
        res = ProcessingResult()
        if not os.path.exists(dst_dir):
            os.makedirs(dst_dir, exist_ok=True)
        shutil.copy2(src_path, dst_path)

        res.set_attribute('hipfile', dst_path)
        if context.param_value('save orig'):
            attr_name = context.param_value('mask hip path attribute')
            res.set_attribute(attr_name, src_path)
        return res

    def postprocess_task(self, context) -> ProcessingResult:
        return ProcessingResult()



import sys
import sqlite3
import argparse
from .config import get_config
from . import paths


def purge_db(db_path):
    with sqlite3.connect(db_path) as con:
        con.executescript('''
UPDATE tasks SET dead=dead|2 WHERE NOT EXISTS (SELECT "group" FROM task_groups WHERE task_id==tasks."id");

PRAGMA foreign_keys=off;
DELETE FROM task_splits WHERE EXISTS (SELECT "id" FROM tasks WHERE "id"==task_splits.task_id AND dead IN (2,3));
DELETE FROM invocations WHERE EXISTS (SELECT "id" FROM tasks WHERE "id"==invocations.task_id AND dead IN (2,3));
DELETE FROM task_groups WHERE EXISTS (SELECT "group" FROM task_group_attributes WHERE "group"==task_groups."group" AND "state"==1);
DELETE FROM task_group_attributes WHERE "state"==1;

DELETE FROM tasks WHERE dead IN (2,3);
PRAGMA foreign_keys=on;
PRAGMA foreign_key_check;
VACUUM;
        ''')
        con.commit()


def info_db(db_path):
    with sqlite3.connect(db_path) as con:
        cur = con.cursor()
        cur.execute('SELECT count("id") FROM tasks')
        task_cnt = cur.fetchone()[0]
        cur.execute('SELECT count("id") FROM invocations')
        invoc_cnt = cur.fetchone()[0]
        cur.execute('SELECT count("id") FROM workers')
        worker_cnt = cur.fetchone()[0]
        cur.execute('SELECT count("group") FROM task_group_attributes')
        groups_cnt = cur.fetchone()[0]

    return {'workers count': worker_cnt,
            'tasks count': task_cnt,
            'invocations count': invoc_cnt,
            'groups count': groups_cnt}


def _main(argv):
    def confirm(prompt='continue? (y/n): '):
        confirm = None
        while confirm not in ('y', 'n'):
            confirm = input(prompt)
            confirm = confirm.lower()

        return confirm == 'y'

    parser = argparse.ArgumentParser(description='scheduler and database maintenance tools')
    parser.add_argument('-y', '--yes', help='do not ask for user confirmation for important operations. useful for scripting', action='store_true')
    t_parsergroup = parser.add_subparsers(dest='target', required=True)

    t_parser = t_parsergroup.add_parser('db', help='perform actions on scheduler database')
    c_parser = t_parsergroup.add_parser('config', help='perform actions on scheduler configuration')

    t_db_parsergroup = t_parser.add_subparsers(dest='db_command', required=True)
    t_db_info_parser = t_db_parsergroup.add_parser('info', help='print some db information')
    t_db_info_parser.add_argument('db_path', nargs='?')
    t_db_purge_parser = t_db_parsergroup.add_parser('purge', help='actually deletes all archived groups, tasks and invocations. this cannot be undone')
    t_db_purge_parser.add_argument('db_path', nargs='?')

    args = parser.parse_args(argv[1:])

    config = get_config('scheduler')
    if args.target == 'db':
        if args.db_command == 'purge':
            db_path = args.db_path if args.db_path is not None else config.get_option_noasync('scheduler.db_path', str(paths.default_main_database_location()))
            if not args.yes:
                print(f'about to purge database at {db_path}')
                print('this operation may take a lot of time, during which DB will be locked, so it is advised not to run it on a live database')
                print('this operation actually deletes all archived groups, tasks and invocations. this cannot be undone')
                if not confirm():
                    return
            print(f'purging db at {db_path}')
            purge_db(db_path)
            print(f'purge completed')
        if args.db_command == 'info':
            db_path = args.db_path if args.db_path is not None else config.get_option_noasync('scheduler.db_path', str(paths.default_main_database_location()))
            print(f'some basic information on db at {db_path}')
            info = info_db(db_path)
            for k, v in info.items():
                print(f'{k}: {v}')
            return





if __name__ == '__main__':
    sys.exit(_main(sys.argv))



from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult
from lifeblood.taskspawn import TaskSpawn
from lifeblood.exceptions import NodeNotReadyToProcess
from lifeblood.enums import NodeParameterType
from lifeblood.uidata import NodeUi

from typing import Iterable


def node_class():
    return RenameAttributes


class RenameAttributes(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'rename attributes'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'rename', 'attribute', 'core'

    @classmethod
    def type_name(cls) -> str:
        return 'rename_attrib'

    def __init__(self, name: str):
        super(RenameAttributes, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            with ui.multigroup_parameter_block('num'):
                with ui.parameters_on_same_line_block():
                    ui.add_parameter('oldname', '<from / to>', NodeParameterType.STRING, 'from')
                    ui.add_parameter('newname', None, NodeParameterType.STRING, 'to')

    def process_task(self, context) -> ProcessingResult:
        res = ProcessingResult()
        attrs = dict(context.task_attributes())
        for i in range(context.param_value('num')):
            attr_oldname = context.param_value(f'oldname_{i}')

            if attr_oldname not in attrs:
                continue

            attr_newname = context.param_value(f'newname_{i}')
            if attr_newname == attr_oldname:
                continue

            res.set_attribute(attr_newname, attrs[attr_oldname])
            res.remove_attribute(attr_oldname)
            attrs[attr_newname] = attrs[attr_oldname]
            del attrs[attr_oldname]

        return res

    def postprocess_task(self, context) -> ProcessingResult:
        return ProcessingResult()



import asyncio
import uuid
import struct

from ..logging import get_logger
from ..stream_wrappers import MessageReceiveStream
from ..messages import Message
from ..queue import MessageQueue
from ..address import DirectAddress
from ..exceptions import MessageReceivingError, NoMessageError, MessageTransferError, MessageTransferTimeoutError
from ..interfaces import MessageStreamFactory

from typing import Callable, Awaitable, Tuple


class IProtocolInstanceCounter:
    def _protocol_inc_count(self, instance):
        raise NotImplementedError()

    def _protocol_dec_count(self, instance):
        raise NotImplementedError()

    def _allowed_new_instances(self) -> bool:
        raise NotImplementedError()


class MessageProtocol(asyncio.StreamReaderProtocol):
    def __init__(self,
                 instance_counter: IProtocolInstanceCounter,
                 reply_address: Tuple[str, int],
                 message_processor_callback: Callable[[Message], Awaitable[bool]]):
        self.__reader = asyncio.StreamReader()
        self.__logger = get_logger('message_protocol')
        self.__callback = message_processor_callback
        self.__listening_address = reply_address
        self.__saved_references = []
        self.__protocol_instance_counter = instance_counter
        self.__needs_to_stop = asyncio.Event()
        super(MessageProtocol, self).__init__(self.__reader, self.connection_callback)

    def stop(self):
        self.__needs_to_stop.set()

    @staticmethod
    async def read_string(reader: asyncio.StreamReader):
        data_size, = struct.unpack('>Q', await reader.readexactly(8))
        return (await reader.readexactly(data_size)).decode('UTF-8')

    async def connection_callback(self, reader: asyncio.StreamReader, writer: asyncio.StreamWriter):
        # there is a bug in py <=3.8, callback task can be GCd
        # see https://bugs.python.org/issue46309
        # so we HAVE to save a reference to self somewhere
        self.__saved_references.append(asyncio.current_task())

        if not self.__protocol_instance_counter._allowed_new_instances():
            self.__logger.warning('new connection opened, but server closing, dropping connection')
            writer.close()
            await writer.wait_closed()
            return

        stop_waiter = asyncio.create_task(self.__needs_to_stop.wait())

        self.__protocol_instance_counter._protocol_inc_count(self)
        try:
            message_stream = None
            potentially_pending_tasks = []
            # first what's sent is return address
            try:
                other_stream_source = await self.read_string(reader)

                message_stream = MessageReceiveStream(reader, writer,
                                                      this_address=DirectAddress(':'.join(str(x) for x in self.__listening_address)),
                                                      other_end_address=DirectAddress(other_stream_source))

                while not reader.at_eof():
                    try:
                        message_waiter = asyncio.create_task(message_stream.receive_data_message())
                        potentially_pending_tasks = [message_waiter, stop_waiter]
                        done, pending = await asyncio.wait(potentially_pending_tasks, return_when=asyncio.FIRST_COMPLETED)
                        if stop_waiter in done:
                            self.__logger.debug('explicitly asked to stop')
                            message_waiter.cancel()
                            stop_waiter = None
                            break
                        message = await message_waiter
                    except NoMessageError:  # either reader is closed, or chain broke, so timeout happened
                        continue
                    except MessageTransferTimeoutError as e:  # partial message received. either reader is closed, or chain broke, so timeout happened
                        self.__logger.warning(f'partial message received, discarding, ignoring ({str(e)})')
                        # we cannot trust this stream anymore,
                        # next attempted read will most likely be some partial crap too
                        break

                    success = False
                    try:
                        success = await self.__callback(message)
                    finally:
                        await message_stream.acknowledge_received_message(success)
            except MessageTransferError as mre:
                e = mre.wrapped_exception()
                if isinstance(e, asyncio.exceptions.IncompleteReadError):
                    if len(e.partial) == 0:
                        self.__logger.debug('read 0 bytes, connection closed')
                    else:
                        self.__logger.error(f'read incomplete {len(e.partial)} bytes')
                elif isinstance(e, asyncio.exceptions.TimeoutError):
                    self.__logger.warning(f'connection timeout happened')
                elif isinstance(e, ConnectionResetError):
                    self.__logger.error('connection was reset. disconnected %s', e)
                elif isinstance(e, ConnectionError):
                    self.__logger.error('connection error. disconnected %s', e)
            except Exception as e:
                self.__logger.exception('unknown error. disconnected %s', e)
                raise  # TODO: why raise if noone is catching?
            finally:
                # cleanup pending tasks
                for task in potentially_pending_tasks:
                    if not task.done():
                        task.cancel()

                if message_stream is not None:
                    message_stream.close()
                    try:
                        await message_stream.wait_closed()
                    except Exception as e:
                        self.__logger.warning(f'failed to close stream: {str(e)}')

                # and just in case:
                if not writer.is_closing():
                    self.__logger.warning('stream did not close underlying writer, attempting to close')
                    writer.close()
                    try:
                        await writer.wait_closed()
                    except Exception as e:
                        self.__logger.warning(f'failed to close stream: {str(e)}')
        finally:
            self.__protocol_instance_counter._protocol_dec_count(self)
            # according to the note in the beginning of the function - now reference can be cleared
            self.__saved_references.remove(asyncio.current_task())



import socket

from lifeblood.basenode import BaseNode
from lifeblood.nodethings import ProcessingResult, InvocationJob
from lifeblood.invocationjob import InvocationRequirements
from lifeblood.enums import NodeParameterType, WorkerType
from lifeblood.taskspawn import TaskSpawn

from typing import Iterable


def node_class():
    return HoudiniDistributedTrackerStopper


class HoudiniDistributedTrackerStopper(BaseNode):
    @classmethod
    def label(cls) -> str:
        return 'houdini sim tracker stopper'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'houdini', 'tracker', 'distributed', 'stopper'

    @classmethod
    def type_name(cls) -> str:
        return 'houdini_distributed_tracker_stopper'

    def __init__(self, name: str):
        super(HoudiniDistributedTrackerStopper, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.5, 0.25, 0.125)
            ui.add_parameter('host', 'sim tracker host', NodeParameterType.STRING, "`task['SIMTRACKER_HOST']`")
            ui.add_parameter('kport', 'kill port', NodeParameterType.INT, 19377)  # TODO: make default to be expression returning attrib kill port
            #ui.add_parameter('use helper', 'run on scheduler helper', NodeParameterType.BOOL, True)

    def process_task(self, context) -> ProcessingResult:
        addr = (context.param_value('host'), context.param_value('kport'))
        s = socket.socket()
        s.connect(addr)
        try:
            s.sendall(b'\0')
        finally:
            s.close()

        res = ProcessingResult()
        return res



from lifeblood.basenode import BaseNodeWithTaskRequirements
from lifeblood.enums import NodeParameterType
from lifeblood.nodethings import ProcessingResult, ProcessingError
from lifeblood.invocationjob import InvocationJob, InvocationEnvironment

from typing import Iterable


description = '''render USD file with houdini's husk (hydra delegate selector)

usd file path: path to USD file to render
output image file path: path of final image. If AOVs are set up - their render location is not affected by this parameter
skip if result already exists: skip rendering if file defined by "output image file path" already exists
'''


def node_class():
    return Husk


class Husk(BaseNodeWithTaskRequirements):
    @classmethod
    def label(cls) -> str:
        return 'husk'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'houdini', 'karma', 'husk', 'usd', 'stock'

    @classmethod
    def type_name(cls) -> str:
        return 'houdini_husk'

    @classmethod
    def description(cls) -> str:
        return description

    def __init__(self, name):
        super(Husk, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.color_scheme().set_main_color(0.5, 0.25, 0.125)
            ui.add_parameter('delegate', 'usd delegate', NodeParameterType.STRING, 'karma')
            ui.add_parameter('delegate options', 'delegate-specific options', NodeParameterType.STRING, '')
            ui.add_separator()
            ui.add_parameter('usd path', 'usd file path', NodeParameterType.STRING, "`task['file']`")
            ui.add_parameter('image path', 'output image file path', NodeParameterType.STRING, "`task['outimage']`")
            ui.add_parameter('skip if exists', 'skip if result already exists', NodeParameterType.BOOL, False)

            ui.parameter('worker type').set_hidden(True)
            ui.parameter('worker type').set_locked(True)

    def process_task(self, context) -> ProcessingResult:
        args = context.task_attributes()

        env = InvocationEnvironment()
        delegate_options = context.param_value('delegate options').strip()

        if context.param_value('skip if exists'):
            script = 'import os\n' \
                     'if not os.path.exists({imgpath}):\n' \
                     '    import sys\n' \
                     '    from subprocess import Popen\n' \
                     "    sys.exit(Popen(['husk', '-V', '2a', '--renderer', {renderer},{doptions} '--make-output-path',{doframe} '-o', {imgpath}, {usdpath}]).wait())\n" \
                     "else:\n" \
                     "    print('image file already exists, skipping work')\n" \
                    .format(imgpath=repr(context.param_value('image path')),
                            usdpath=repr(context.param_value('usd path')),
                            doframe=f" '-f', {repr(str(args['frames'][0]))}," if 'frames' in args else '',
                            renderer=repr(context.param_value('delegate')),
                            doptions=f' "--delegate-options", {repr(delegate_options)},' if delegate_options else ''
                            )

            invoc = InvocationJob(['python', ':/karmacall.py'])
            invoc.set_extra_file('karmacall.py', script)
        else:  # TODO: -f there is testing, if succ - make a parameter out of it on the node or smth
            invoc = InvocationJob(['husk', '-V', '2a',
                                   '--renderer', context.param_value('delegate')] +
                                  (['--delegate-options', delegate_options] if delegate_options else []) +
                                  ['--make-output-path'] +
                                  (['-f', str(args['frames'][0])] if 'frames' in args else []) +
                                  ['-o', context.param_value('image path'), context.param_value('usd path')],
                                  env=env)
        res = ProcessingResult(invoc)
        return res

    def postprocess_task(self, context) -> ProcessingResult:
        res = ProcessingResult()
        res.set_attribute('file', context.param_value('image path'))
        return res



# import asyncio
# from .stream_wrappers import MessageStream
# from .address import split_address
#
# from typing import Tuple, Union
#
#
# async def open_sending_stream(address: Union[str, Tuple[Tuple[str, int], ...]], reply_address: Tuple[str, int]):
#     """
#     address is expected to be in form of "1.2.3.4:1313|2.3.4.5:2424"...
#     """
#     if isinstance(address, str):
#         address = split_address(address)
#     reader, writer = await asyncio.open_connection(address[0][0], address[0][1])
#
#     return MessageStream(reader, writer, reply_address)



from lifeblood.basenode import ProcessingResult, ProcessingContext, ProcessingError
from lifeblood.invocationjob import InvocationJob
from lifeblood.enums import NodeParameterType
from lifeblood_alicevision_modules.base_node import AlicevisionBaseNode
from pathlib import Path
from typing import Iterable


def node_class():
    return AlicevisionStructureFromMotion


class AlicevisionStructureFromMotion(AlicevisionBaseNode):
    def __init__(self, name):
        super(AlicevisionStructureFromMotion, self).__init__(name)
        ui = self.get_ui()
        with ui.initializing_interface_lock():
            ui.add_parameter('input', 'Input Images Directory', NodeParameterType.STRING, "`task['av_camera_sfm']`")
            ui.add_parameter('featuresFolders', 'Features Dir', NodeParameterType.STRING, "`task['av_features_dir']`")  # TODO: this seem to be an array of dirs, so add support
            ui.add_parameter('matchesFolders', 'Matches Dir', NodeParameterType.STRING, "`task['av_feature_matches']`")  # TODO: this seem to be an array of dirs, so add support
            ui.add_parameter('output', 'Output Structure From Motion', NodeParameterType.STRING, "`config['global_scratch_location']`/alicevision/`task['uuid']`/StructureFromMotion/sfm.abc")
            ui.add_parameter('outputViewsAndPoses', 'Output Views And Poses', NodeParameterType.STRING, "`config['global_scratch_location']`/alicevision/`task['uuid']`/StructureFromMotion/cameras.sfm")
            ui.add_parameter('extraInfoFolder', 'Output Folder', NodeParameterType.STRING, "`config['global_scratch_location']`/alicevision/`task['uuid']`/StructureFromMotion")
            ui.add_separator()
            ui.add_parameter('describerTypes', 'Describer Types', NodeParameterType.STRING, 'sift')  # TODO: make a set of checkboxes
            ui.add_parameter('lockScenePreviouslyReconstructed', 'Lock Scene Previously Reconstructed', NodeParameterType.BOOL, False)
            ui.add_parameter('useLocalBA', 'Local Bundle Adjustment', NodeParameterType.BOOL, True)
            ui.add_parameter('maxNumberOfMatches', 'Max Number Of Matches', NodeParameterType.INT, 0)
            ui.add_parameter('minNumberOfMatches', 'Min Number Of Matches', NodeParameterType.INT, 0)
            ui.add_parameter('minInputTrackLength', 'Min Input Track Length', NodeParameterType.INT, 2).set_value_limits(2)
            ui.add_parameter('lockAllIntrinsics', 'Force Lock of All Intrinsic Camera Parameters', NodeParameterType.BOOL, False)
            ui.add_parameter('filterTrackForks', 'Filter Track Forks', NodeParameterType.BOOL, False)
            with ui.parameters_on_same_line_block():
                ui.add_parameter('initialPairA', 'Initial Pair', NodeParameterType.STRING, '')
                ui.add_parameter('initialPairB', None, NodeParameterType.STRING, '')


    @classmethod
    def label(cls) -> str:
        return 'AV Structure From Motion'

    @classmethod
    def tags(cls) -> Iterable[str]:
        return 'alicevision', 'structure', 'motion', 'sfm'

    @classmethod
    def type_name(cls) -> str:
        return 'stock_alicevision_structure_from_motion'

    @classmethod
    def description(cls) -> str:
        return 'this node type does not have a description'

    def process_task(self, context: ProcessingContext) -> ProcessingResult:
        output = Path(context.param_value('output').strip())
        if not output.is_absolute():
            raise ProcessingError('output path must be absolute and not empty')
        output_vap = Path(context.param_value('outputViewsAndPoses'))
        if not output_vap.is_absolute():
            raise ProcessingError('output path must be absolute and not empty')
        output_folder = Path(context.param_value('extraInfoFolder'))
        if not output_folder.is_absolute():
            raise ProcessingError('output path must be absolute and not empty')

        output.parent.mkdir(parents=True, exist_ok=True)  # TODO: move this to worker
        output_vap.parent.mkdir(parents=True, exist_ok=True)  # TODO: move this to worker
        output_folder.mkdir(parents=True, exist_ok=True)  # TODO: move this to worker

        args = ['aliceVision_incrementalSfM', '--verboseLevel', 'info']
        for param_name in ('input', 'featuresFolders', 'matchesFolders',
                           'describerTypes', 'lockScenePreviouslyReconstructed', 'useLocalBA',
                           'maxNumberOfMatches', 'minNumberOfMatches', 'minInputTrackLength',
                           'lockAllIntrinsics', 'filterTrackForks', 'initialPairA', 'initialPairB'):
            args += [f'--{param_name}', context.param_value(param_name)]
        args += ['--output', output]
        args += ['--outputViewsAndPoses', output_vap]
        args += ['--extraInfoFolder', output_folder]

        job = InvocationJob([str(x) for x in args])
        res = ProcessingResult(job)
        res.set_attribute('av_structure_from_motion', str(output))
        res.set_attribute('av_viewes_and_poses', str(output_vap))
        res.set_attribute('av_extra_info_folder', str(output_folder))
        return res
